package net.refractions.udig.catalog;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

import net.refractions.udig.catalog.internal.CatalogImpl;
import net.refractions.udig.catalog.internal.Policy;
import net.refractions.udig.catalog.internal.ServiceFactoryImpl;
import net.refractions.udig.core.internal.ExtensionPointProcessor;
import net.refractions.udig.core.internal.ExtensionPointUtil;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.preferences.IEclipsePreferences;
import org.eclipse.core.runtime.preferences.IPreferencesService;
import org.eclipse.core.runtime.preferences.InstanceScope;
import org.osgi.framework.BundleContext;
import org.osgi.service.prefs.BackingStoreException;
import org.osgi.service.prefs.Preferences;

/**
 * The main plugin class to be used in the desktop.
 */
public class CatalogPlugin extends Plugin {
    public static final String ID = "net.refractions.udig.catalog"; //$NON-NLS-1$

    // The shared instance.
    private static CatalogPlugin plugin;

    // Resource bundle.
    private ResourceBundle resourceBundle;

    /**
     * The constructor.
     */
    public CatalogPlugin() {
        super();
        plugin = this;
        catalogs = new ICatalog[]{new CatalogImpl()};
        serviceFactory = new ServiceFactoryImpl();
    }

    /**
     * This method is called upon plug-in activation
     */
    public void start( BundleContext context ) throws Exception {
        super.start(context);
        Policy.localize("net.refractions.udig.catalog.messages"); //$NON-NLS-1$
        getDefault();
    }
    /**
     * Cleanup after shared images.
     * 
     * @see org.eclipse.ui.plugin.AbstractUIPlugin#stop(org.osgi.framework.BundleContext)
     * @param context
     * @throws Exception
     */
    public void stop( BundleContext context ) throws Exception {
        plugin = null;
        resourceBundle = null;

        storeToPreferences();
        Policy.cleanup();
        super.stop(context);
    }

    public void restoreFromPreferences() throws BackingStoreException, MalformedURLException {
        try {
            IPreferencesService prefs = Platform.getPreferencesService();
            IEclipsePreferences root = prefs.getRootNode();
            Preferences node = root.node(InstanceScope.SCOPE).node(ID + ".services"); //$NON-NLS-1$

            for( String id : node.childrenNames() ) {
                URL url;
                try {
                    url = new URL(URLDecoder.decode(id, "UTF-8")); //$NON-NLS-1$
                } catch (UnsupportedEncodingException e) {
                    CatalogPlugin.log( "Could not code preferences URL", e ); //$NON-NLS-1$
                    throw new MalformedURLException(e.toString());
                }
                Preferences service = node.node(id);

                String[] keys = service.keys();
                Map<String, Serializable> map = new HashMap<String, Serializable>();
                for( int j = 0; j < keys.length; j++ ) {
                    map.put(keys[j], service.get(keys[j], null));
                }
                for( IService serv : serviceFactory.aquire(url, map) ) {
                    catalogs[0].add(serv);
                }
            }

            loadCatalogs();

        } catch (BackingStoreException t) {
            CatalogPlugin.log(null, t);
            throw t;
        } catch (MalformedURLException t) {
            CatalogPlugin.log(null, t);
            throw t;
        } catch (Throwable t) {
            CatalogPlugin.log(null, new Exception(t));
        }
    }

    private void loadCatalogs() {
        final List<ICatalog> cats = new LinkedList<ICatalog>();
        ExtensionPointUtil.process(getDefault(),
                "net.refractions.udig.catalog.ICatalog", new ExtensionPointProcessor(){ //$NON-NLS-1$
                    public void process( IExtension extension, IConfigurationElement element )
                            throws Exception {
                        cats.add((ICatalog) element.createExecutableExtension("class")); //$NON-NLS-1$                 
                    }
                });

        ICatalog[] ctemp = new ICatalog[cats.size() + 1];
        ctemp[0] = catalogs[0];
        Iterator<ICatalog> ii = cats.iterator();
        for( int i = 1; i < ctemp.length && ii.hasNext(); i++ ) {
            ctemp[i] = ii.next();
        }
        catalogs = ctemp;
    }

    public void storeToPreferences() throws BackingStoreException, IOException {
        IPreferencesService prefs = Platform.getPreferencesService();
        IEclipsePreferences root = prefs.getRootNode();
        Preferences node = root.node(InstanceScope.SCOPE).node(ID + ".services"); //$NON-NLS-1$
        clearPreferences(node);
        for( IResolve member : getLocalCatalog().members(null) ) {
            try {
                IService service = (IService) member;
                String id;
                try {
                    id = URLEncoder.encode(service.getIdentifier().toString(), "UTF-8"); //$NON-NLS-1$
                } catch (UnsupportedEncodingException e1) {
                    // should never happen
                    CatalogPlugin.log(null, e1);
                    throw new BackingStoreException(e1.toString());
                }

                Preferences serviceNode = node.node(id);

                for( Iterator e = service.getConnectionParams().entrySet().iterator(); e.hasNext(); ) {
                    @SuppressWarnings("all") //$NON-NLS-1$
                    Map.Entry entry = (Map.Entry) e.next();
                    @SuppressWarnings("all") //$NON-NLS-1$
                    String key = entry.getKey().toString();

                    @SuppressWarnings("all") //$NON-NLS-1$
                    String value = entry.getValue() == null ? null : entry.getValue().toString();

                    if (value != null)
                        serviceNode.put(key, value);
                }
                if (serviceNode.keys().length > 0)
                    serviceNode.flush();
            } catch (Exception e) {
                // continue
            }
        }
        node.flush();
    }

    private void clearPreferences( Preferences node ) throws BackingStoreException {
        for( String name : node.childrenNames() ) {
            Preferences child = node.node(name);
            child.removeNode();
        }
    }

    private ICatalog[] catalogs;
    private IServiceFactory serviceFactory;

    private static Boolean loading = Boolean.FALSE;
    /**
     * Returns the shared instance.
     */
    public static CatalogPlugin getDefault() {
        if (plugin == null || !plugin.loaded)
            synchronized (CatalogPlugin.class) {
                try {
                    if (!plugin.loaded && !loading.booleanValue()) {
                        loading = Boolean.TRUE;
                        plugin.restoreFromPreferences();
                        loading = Boolean.FALSE;
                    }
                } catch (BackingStoreException e) {
                    CatalogPlugin.log( null, e);
                    plugin.loaded = false;
                    loading = Boolean.FALSE;
                    // nothing ... not loaded
                } catch (MalformedURLException e) {
                    CatalogPlugin.log( null, e);
                    plugin.loaded = false;
                    loading = Boolean.FALSE;
                    // nothing ... not loaded
                }
                plugin.loaded = true;
            }
        return plugin;
    }
    private boolean loaded = false;

    /**
     * Add a catalog listener for changed to this catalog.
     * 
     * @param listener
     */
    public static void addListener( IResolveChangeListener listener ) {
        ((CatalogImpl) getDefault().getLocalCatalog()).addCatalogListener(listener);
    }

    /**
     * Remove a catalog listener that was interested in this catalog.
     * 
     * @param listener
     */
    public static void removeListener( IResolveChangeListener listener ) {
        ((CatalogImpl) getDefault().getLocalCatalog()).removeCatalogListener(listener);
    }

    /**
     * Returns the string from the plugin's resource bundle, or 'key' if not found.
     */
    public static String getResourceString( String key ) {
        ResourceBundle bundle = CatalogPlugin.getDefault().getResourceBundle();
        try {
            return (bundle != null) ? bundle.getString(key) : key;
        } catch (MissingResourceException e) {
            return key;
        }
    }

    /**
     * Returns the plugin's resource bundle,
     */
    public ResourceBundle getResourceBundle() {
        try {
            if (resourceBundle == null)
                resourceBundle = ResourceBundle
                        .getBundle("net.refractions.udig.catalog.CatalogPluginResources"); //$NON-NLS-1$
        } catch (MissingResourceException x) {
            resourceBundle = null;
        }
        return resourceBundle;
    }
    /**
     * @return Returns the catalogs found ... local one is slot[0]. TODO hook the extension point
     *         up.
     */
    public ICatalog[] getCatalogs() {
        if (catalogs == null)
            loadCatalogs();
        return catalogs;
    }

    /**
     * @return the local catalog. Equivalent to getCatalogs()[0]
     */
    public ICatalog getLocalCatalog() {
        return getCatalogs()[0];
    }
    /**
     * @return Returns the serviceFactory.
     */
    public IServiceFactory getServiceFactory() {
        return serviceFactory;
    }

    /**
     * Attempts to turn data into a URL. If data is an array or Collection, it will return the first
     * successful URL it can find. This is a utility method. Feel free to move it to another class.
     * In the future, it might be nice to have it return a List of the URLs it found, not just the
     * first one.
     * 
     * @param data
     * @return a URL if it can find one, or null otherwise
     */
    public static URL locateURL( Object data ) {
        if (data == null)
            return null;

        return toURL(data);

        // if (data instanceof Collection) {
        // for (Object item : (Collection) data) {
        // URL url = toURL(item);
        // if (url != null) {
        // return url;
        // }
        // }
        // }
        // else {
        // return toURL(data);
        // }

        // return null;
    }

    private static URL toURL( Object data ) {
        URL url = null;
        if (data instanceof String) {
            String string = (String) data;
            int index = string.indexOf("\n"); //$NON-NLS-1$
            if (index > -1)
                string = string.substring(0, index);

            // try to turn into a string directly
            try {
                url = new URL(string);
            } catch (MalformedURLException e) {
                // try to go to a file first
                try {
                    url = new File(string).toURL();
                } catch (MalformedURLException e1) {
                    // do nothing
                }
            }

        } else if (data instanceof File) {
            try {
                url = ((File) data).toURL();
            } catch (MalformedURLException e) {
                // do nothing
            }
        } else if (data instanceof URL) {
            url = (URL) data;
        } else if (data instanceof IGeoResource) {
            IGeoResource resource = (IGeoResource) data;
            return resource.getIdentifier();
        } else if (data instanceof IService) {
            IService service = (IService) data;
            return service.getIdentifier();
        }

        return url;
    }
    
    /**
     * Logs the Throwable in the plugin's log.
     * <p>
     * This will be a user visable ERROR iff:
     * <ul>
     * <li>t is an Exception we are assuming it is human readable or if a message is provided
     */
    public static void log( String message, Throwable t ) {
    	if( message==null )
    	message=""; //$NON-NLS-1$
        int status = t instanceof Exception || message != null ? IStatus.ERROR : IStatus.WARNING;
        getDefault().getLog().log(new Status(status, ID, IStatus.OK, message, t));
    }
    /**
     * Messages that only engage if getDefault().isDebugging()
     * <p>
     * It is much prefered to do this:<pre><code>
     * private static final String RENDERING = "net.refractions.udig.project/render/trace";
     * if( ProjectUIPlugin.getDefault().isDebugging() && "true".equalsIgnoreCase( RENDERING ) ){
     *      System.out.println( "your message here" );
     * }
     */
    public static void trace( String message, Throwable e) {
        if( getDefault().isDebugging() ) {
            if( message != null ) System.out.println( message );
            if( e != null ) e.printStackTrace();
        }
    }
    /**
     * Performs the Platform.getDebugOption true check on the provided trace
     * <p>
     * Note: ProjectUIPlugin.getDefault().isDebugging() must also be on.
     * <ul>
     * <li>Trace.RENDER - trace rendering progress
     * </ul>
     * </p> 
     * @param trace currently only RENDER is defined
     */
    public static boolean isDebugging( final String trace ){
        return getDefault().isDebugging() &&
            "true".equalsIgnoreCase(Platform.getDebugOption(trace)); //$NON-NLS-1$    
    }
}
