/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.catalog;

import java.io.IOException;
import java.io.Serializable;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IProgressMonitor;


/**
 * 
 * Represents a geo spatial service handle. Follows the same design as IResource.
 * <p>
 * Represents a spatial service, which may be lazily loaded. The existance of this object does not 
 * ensure that the advertized data is guaranteed to exist, nor does this interface guarantee that 
 * the service exists based on this object's existance. We should also note the resource management 
 * is left to the user, and that resolve() is not guaranteed to return the same instance object 
 * from two subsequent calls, but may. This is merely a handle to some information about a service, 
 * and a method of aquiring an instance of the service ... 
 * </p>
 * <p>
 * NOTE: This may be the result of communications with a metadata service, and as such this service 
 * handle may not have been validated yet. Remember to check the service status.
 * </p>
 * @author David Zwiers, Refractions Research
 * @since 0.6
 * 
 * @see IServiceInfo
 * @see IServiceFactory
 */
public abstract class IService implements IResolve {    
    /**
     * 
     * Will attempt to morph into the adaptee, and return that object.
     * 
     * Required adaptions:
     * <ul>
     * <li>IServiceInfo.class
     * <li>List.class <IGeoResource>
     * </ul>
     * 
     * May Block.
     * 
     * @param adaptee
     * @param monitor
     * @return instance of adaptee, or null if unavailable (IServiceInfo and List<IGeoResource> must be supported)
     * 
     * @see IServiceInfo
     * @see IGeoResource
     * @see IResolve#resolve(Class, IProgressMonitor)
     */
    public abstract <T> T resolve(Class<T> adaptee,IProgressMonitor monitor) throws IOException;
    
    /**
     * Returns LocalCatalog by defaul, subclass must override iff
     * a custom catalog is used.
     */
    public ICatalog parent( IProgressMonitor monitor ) {
        return CatalogPlugin.getDefault().getLocalCatalog();
    }
    /**
     * Return list of IGeoResources managed by this service.
     * <p>
     * Many file based serivces will just contain a single IGeoResource.
     * </p>
     * 
     */
    public abstract List<? extends IGeoResource> members( IProgressMonitor monitor ) throws IOException;
    
    /**
     * @return IServiceInfo resolve(IServiceInfo.class,IProgressMonitor monitor);
     * 
     * @see IService#resolve(Class, IProgressMonitor)
     */
    public IServiceInfo getInfo(IProgressMonitor monitor) throws IOException{
        return resolve(IServiceInfo.class,monitor);
    }
    /**
     *
     * Accessor to the set of params used to create this entry. There is no guarantee that these 
     * params created a usable service (@see getStatus() ). These params may have been modified 
     * within the factory during creation. This method is intended to be used for cloning 
     * (@see IServiceFactory) or for persistence between sessions. 
     * 
     * @see IServiceFactory
     * 
     * @return
     */
    public abstract Map<String,Serializable> getConnectionParams();
    
    /**
     * 
     * This should represent the identifier
     * 
     * @see Object#equals(java.lang.Object)
     * 
     * @param arg0
     * @return
     */
    public final boolean equals( Object arg0 ){
        if(arg0!=null && arg0 instanceof IService){
            IService service = (IService)arg0;
            if(getIdentifier()!=null && service.getIdentifier()!=null)
                return getIdentifier().equals(service.getIdentifier());
        }
        return false;
    }
    /**
     * 
     * This should represent the identified
     * 
     * @see Object#hashCode()
     * 
     * @return
     */
    public final int hashCode(){
        if(getIdentifier()!=null)
            return getIdentifier().hashCode();
        return super.hashCode();
    }
    
    /**
     * Indicate class and id.
     * 
     * @return string representing this IResolve
     */
    public String toString(){
        StringBuffer buf = new StringBuffer();
        String classname = getClass().getName();
        String name = classname.substring( classname.lastIndexOf('.')+1);
        buf.append( name );
        buf.append( "(" ); //$NON-NLS-1$
        buf.append( getIdentifier() );
        buf.append( ")" ); //$NON-NLS-1$
        return buf.toString();
    }
}
