/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.catalog.internal.geotiff;

import java.io.File;
import java.io.Serializable;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLDecoder;
import java.util.HashMap;
import java.util.Map;

import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.ServiceExtension;

import org.geotools.gce.geotiff.GeoTiffFormat;
import org.geotools.gce.geotiff.GeoTiffFormatFactorySpi;

/**
 * Provides the interface to the catalog service extension point.
 * <p>
 * This class is responsible for ensuring that only those services that the GeoTiff plugin is
 * capable of processing are created.
 * </p>
 * 
 * @author mleslie
 * @since 0.6.0
 */
public class GeoTiffServiceExtension implements ServiceExtension {
    /** <code>URL_PARAM</code> field */
    public static String URL_PARAM = "URL"; //$NON-NLS-1$

    private static GeoTiffFormatFactorySpi factory;
    private static GeoTiffFormat format;

    /**
     * Construct <code>GeoTiffServiceExtension</code>.
     */
    public GeoTiffServiceExtension() {
        super();
    }

    public IService createService( URL id, Map<String, Serializable> params ) {
        if (id == null) {
            if (params.containsKey(URL_PARAM)) {
                Object param = params.get(URL_PARAM);
                if (param instanceof String) {
                    try {
                        id = new URL((String) param);
                    } catch (MalformedURLException ex) {
                        return null;
                    }
                } else if (param instanceof URL) {
                    id = (URL) param;
                } else {
                    return null;
                }
            } else {
                return null;
            }
        }
        if (!canProcess(id)) {
            return null;
        }
        GeoTiffServiceImpl service = new GeoTiffServiceImpl(id, getFactory());
        return service;
    }

    private static GeoTiffFormat getFormat() {
        if (format == null) {
            format = (GeoTiffFormat) getFactory().createFormat();
        }
        return format;
    }

    /**
     * Finds or creates a GeoTiffFormatFactorySpi.
     * 
     * @return Default GeoTiffFormatFactorySpi
     */
    public static GeoTiffFormatFactorySpi getFactory() {
        if (factory == null) {
            factory = new GeoTiffFormatFactorySpi();
        }
        return factory;
    }

    private boolean canProcess( URL id ) {
        if (id == null) {
            return false;
        }

        File file = null;
        try {
            file = new File(id.toURI());
        } catch (IllegalArgumentException ex) {
            return false;
        } catch (URISyntaxException ex) {
            return false;
        }
        try {
            return getFormat().accepts(file);
        } catch (RuntimeException ex) {
            return false;
        }
    }

    public Map<String, Serializable> createParams( URL url ) {
        if (!canProcess(url))
            return null;

        Map<String, Serializable> params = new HashMap<String, Serializable>();
        if (url != null) {
            params.put(URL_PARAM, url);
        }
        return params;
    }
}
