/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.catalog.internal.shp;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

import net.refractions.udig.catalog.CatalogPlugin;
import net.refractions.udig.catalog.IResolveChangeEvent;
import net.refractions.udig.catalog.IResolveDelta;
import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.IServiceInfo;
import net.refractions.udig.catalog.internal.CatalogImpl;
import net.refractions.udig.catalog.internal.ResolveChangeEvent;
import net.refractions.udig.catalog.internal.ResolveDelta;

import org.eclipse.core.runtime.IProgressMonitor;
import org.geotools.data.DefaultQuery;
import org.geotools.data.Transaction;
import org.geotools.data.shapefile.ShapefileDataStore;
import org.geotools.data.shapefile.indexed.IndexedShapefileDataStoreFactory;

/**
 * Connect to a shapefile
 * 
 * @author David Zwiers, Refractions Research
 * @since 0.6
 */
public class ShpServiceImpl extends IService {

	public static final String SHP_TRACE = "net.refractions.udig.catalog.shp/debug/trace"; //$NON-NLS-1$

	private URL url = null;

	private Map<String, Serializable> params = null;

	/**
	 * Construct <code>ShpServiceImpl</code>.
	 * 
	 * @param arg1
	 * @param arg2
	 */
	public ShpServiceImpl(URL arg1, Map<String, Serializable> arg2) {
		url = arg1;
		params = arg2;
		Serializable memorymapped = params.get("memory mapped buffer"); //$NON-NLS-1$
		if (memorymapped == null) {
			memorymapped = false;
			try {
				File file = new File(url.getFile());
				final int maxsize = 1024 * 2048;
				if (file.length() > maxsize) {
					memorymapped = false;
				}
			} catch (Exception e) {
				memorymapped = false;
			}
			params.put(IndexedShapefileDataStoreFactory.MEMORY_MAPPED.key,
					memorymapped); //$NON-NLS-1$
		}
		params.put(IndexedShapefileDataStoreFactory.CREATE_SPATIAL_INDEX.key,
				true);
		params.put(IndexedShapefileDataStoreFactory.SPATIAL_INDEX_TYPE.key,
				IndexedShapefileDataStoreFactory.TREE_QIX);
	}

	/*
	 * Required adaptions: <ul> <li>IServiceInfo.class <li>List.class
	 * <IGeoResource> </ul>
	 * 
	 * @see net.refractions.udig.catalog.IService#resolve(java.lang.Class,
	 *      org.eclipse.core.runtime.IProgressMonitor)
	 */
	public <T> T resolve(Class<T> adaptee, IProgressMonitor monitor)
			throws IOException {
		if (adaptee == null)
			return null;
		if (adaptee.isAssignableFrom(IServiceInfo.class))
			return adaptee.cast(this.getInfo(monitor));
		if (adaptee.isAssignableFrom(List.class))
			return adaptee.cast(members(monitor));
		if (adaptee.isAssignableFrom(ShapefileDataStore.class))
			return adaptee.cast(getDS());
		return null;
	}

	/*
	 * @see net.refractions.udig.catalog.IResolve#canResolve(java.lang.Class)
	 */
	public <T> boolean canResolve(Class<T> adaptee) {
		if (adaptee == null)
			return false;
		return (adaptee.isAssignableFrom(IServiceInfo.class)
				|| adaptee.isAssignableFrom(List.class) || adaptee
				.isAssignableFrom(ShapefileDataStore.class));
	}

	/*
	 * @see net.refractions.udig.catalog.IResolve#members(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public List<ShpGeoResourceImpl> members(IProgressMonitor monitor)
			throws IOException {

		if (members == null) {
			synchronized (getDS()) {
				if (members == null) {
					members = new LinkedList<ShpGeoResourceImpl>();
					String[] typenames = ds.getTypeNames();
					if (typenames != null)
						for (int i = 0; i < typenames.length; i++) {
							members.add(new ShpGeoResourceImpl(this,
									typenames[i]));
						}
				}
			}
		}
		return members;
	}

	private List<ShpGeoResourceImpl> members = null;

	/*
	 * @see net.refractions.udig.catalog.IService#getInfo(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public IServiceInfo getInfo(IProgressMonitor monitor) throws IOException {
		getDS(); // load ds
		if (info == null && ds != null) {
			synchronized (ds) {
				if (info == null) {
					info = new IServiceShpInfo();
				}
			}
			IResolveDelta delta = new ResolveDelta(this,
					IResolveDelta.Kind.CHANGED);
			((CatalogImpl) CatalogPlugin.getDefault().getLocalCatalog())
					.fire(new ResolveChangeEvent(this,
							IResolveChangeEvent.Type.POST_CHANGE, delta));
		}
		return info;
	}

	private IServiceInfo info = null;

	/*
	 * @see net.refractions.udig.catalog.IService#getConnectionParams()
	 */
	public Map<String, Serializable> getConnectionParams() {
		return params;
	}

	private Throwable msg = null;

	ShapefileDataStore ds = null;

	ShapefileDataStore getDS() throws IOException {
		if (ds == null) {
			synchronized (ShapefileDataStore.class) {
				if( ShpPlugin.isDebugging(SHP_TRACE)){
					ClassLoader current = getClass().getClassLoader();
					try{
					Thread.currentThread().setContextClassLoader(ShapefileDataStore.class.getClassLoader());
					Logger logger = Logger.getLogger("org.geotools.data.shapefile");//$NON-NLS-1$
					logger.setLevel(Level.FINEST);
					logger.addHandler(new Handler(){

						@Override
						public void publish(LogRecord record) {
							System.err.println(record.getMessage());
						}

						@Override
						public void flush() {
							System.err.flush();
						}

						@Override
						public void close() throws SecurityException {
							
						} 
						
					});
					}finally{
						Thread.currentThread().setContextClassLoader(current);						
					}
					
				}
				if (ds == null) {
					IndexedShapefileDataStoreFactory dsf = new IndexedShapefileDataStoreFactory();
					if (dsf.canProcess(params)) {
						try {
							ds = (ShapefileDataStore) dsf
									.createDataStore(params);

							// kick it in order to generate index files if
							// necessary.
							ds.getFeatureReader(new DefaultQuery(ds.getSchema()
									.getTypeName()), Transaction.AUTO_COMMIT).close();
						} catch (IOException e) {
							msg = e;
							try {
								params
										.remove(IndexedShapefileDataStoreFactory.CREATE_SPATIAL_INDEX.key);
								ds = (ShapefileDataStore) dsf
										.createDataStore(params);

							} catch (IOException  e2) {
								msg = e2;
								throw e2;
							}
						}
					}
				}
			}
			IResolveDelta delta = new ResolveDelta(this,
					IResolveDelta.Kind.CHANGED);
			((CatalogImpl) CatalogPlugin.getDefault().getLocalCatalog())
					.fire(new ResolveChangeEvent(this,
							IResolveChangeEvent.Type.POST_CHANGE, delta));
		}
		return ds;
	}

	/*
	 * @see net.refractions.udig.catalog.IResolve#getStatus()
	 */
	public Status getStatus() {
		return msg != null ? Status.BROKEN : ds == null ? Status.NOTCONNECTED
				: Status.CONNECTED;
	}

	/*
	 * @see net.refractions.udig.catalog.IResolve#getMessage()
	 */
	public Throwable getMessage() {
		return msg;
	}

	/*
	 * @see net.refractions.udig.catalog.IResolve#getIdentifier()
	 */
	public URL getIdentifier() {
		return url;
	}

	private class IServiceShpInfo extends IServiceInfo {

		IServiceShpInfo() {
			super();
			keywords = new String[] { ".shp", "Shapefile", //$NON-NLS-1$ //$NON-NLS-2$
					ds.getTypeNames()[0] };

			try {
				schema = new URI("shp://www.opengis.net/gml"); //$NON-NLS-1$
			} catch (URISyntaxException e) {
				ShpPlugin.log(null, e);
				schema = null;
			}
		}

		public String getDescription() {
			return getIdentifier().toString();
		}

		public String getTitle() {
			return getIdentifier().getFile();
		}
	}
}
