/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.catalog.internal.worldimage;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.channels.Channels;
import java.nio.channels.ReadableByteChannel;
import java.util.HashMap;

import net.refractions.udig.catalog.CatalogPlugin;
import net.refractions.udig.catalog.IGeoResourceInfo;
import net.refractions.udig.catalog.IResolveChangeEvent;
import net.refractions.udig.catalog.IResolveDelta;
import net.refractions.udig.catalog.internal.CatalogImpl;
import net.refractions.udig.catalog.internal.ResolveChangeEvent;
import net.refractions.udig.catalog.internal.ResolveDelta;
import net.refractions.udig.catalog.rasterings.AbstractRasterGeoResource;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.geotools.data.shapefile.prj.PrjFileReader;
import org.geotools.gce.image.WorldImageFormat;
import org.geotools.geometry.JTS.ReferencedEnvelope;
import org.geotools.parameter.DefaultParameterDescriptor;
import org.geotools.parameter.DefaultParameterDescriptorGroup;
import org.geotools.parameter.ParameterGroup;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.coverage.grid.GridCoverage;
import org.opengis.parameter.GeneralParameterDescriptor;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import com.vividsolutions.jts.geom.Envelope;

/**
 * Provides a handle to a world image resource allowing the service to be lazily 
 * loaded.
 * @author mleslie
 * @since 0.6.0
 */
public class WorldImageGeoResourceImpl extends AbstractRasterGeoResource {
    private IGeoResourceInfo info;
    private String prjName;
    String name;

    /**
     * Construct <code>WorldImageGeoResourceImpl</code>.
     *
     * @param service Service creating this resource.
     * @param name Human readable name of this resource.
     * @param prjName Name a projection file associated with this resource
     *                can be expected to have.
     */
    public WorldImageGeoResourceImpl(WorldImageServiceImpl service, 
            String name, String prjName) {
        super(service, name);
        this.prjName = prjName;
    }
 
    public IGeoResourceInfo getInfo(IProgressMonitor monitor) throws IOException {
        if(this.info == null && getStatus() != Status.BROKEN) {
            this.info = new IGeoResourceWorldImageInfo();
            IResolveDelta delta = new ResolveDelta(this, IResolveDelta.Kind.CHANGED);
            ((CatalogImpl)CatalogPlugin.getDefault().getLocalCatalog()).fire(
                    new ResolveChangeEvent(this, IResolveChangeEvent.Type.POST_CHANGE,
                    delta));
        }
        return this.info;
    }
    
    /**
     * Convenience method for opening a PrjFileReader.  Stolen from Geotools
     * ShapefileSomethingROther
     *
     * @return A new PrjFileReader 
     *
     * @throws IOException If an error occurs during creation.
     */
    private PrjFileReader openPrjReader(URL prjURL) 
            throws IOException, FactoryException {
        ReadableByteChannel rbc = null;
        try{
            rbc = getReadChannel(prjURL);
        }
        catch(IOException e){
            CatalogPlugin.getDefault().getLog().log(
                    new org.eclipse.core.runtime.Status(IStatus.WARNING, 
                    "net.refractions.udig.catalog", 0,  //$NON-NLS-1$
                    Policy.bind("WorldImageGeoResourceImpl.PrjUnavailable"), e )); //$NON-NLS-1$
        }
        if (rbc == null) {
            return null;
        }
        
        return new PrjFileReader(rbc);
    }
    
    /**
     * Convienience method to create a ReadableByteChannel from a URL.
     * 
     * @param url
     * @return A Channel for the given file
     * @throws IOException
     */
    private ReadableByteChannel getReadChannel(URL url)
            throws IOException {
        ReadableByteChannel channel = null;

        if (url.getProtocol().equals("file")) { //$NON-NLS-1$
            File file = null;
            if(url.getHost()!=null && !url.getHost().equals("")) { //$NON-NLS-1$
                //win
                file = new File(url.getHost()+":"+url.getFile()); //$NON-NLS-1$
            }else {
                //linux
                file = new File(url.getFile());
            }

            if (!file.exists() || !file.canRead()) {
            	throw new FileNotFoundException(file.getAbsolutePath());
            }                        
            FileInputStream in = new FileInputStream(file);            
            channel = in.getChannel();
            
        } else {
            InputStream in = url.openConnection().getInputStream();
            channel = Channels.newChannel(in);
        }
        
        return channel;
    }
    
    public ParameterGroup getReadParameters() {
        try {
        	PrjFileReader prjRead;
			try{
        		prjRead = openPrjReader(new URL(this.prjName));
        	}catch(FileNotFoundException e){
        		prjRead = null;
        		 CatalogPlugin.getDefault().getLog().log(
                         new org.eclipse.core.runtime.Status(IStatus.WARNING, 
                         "net.refractions.udig.catalog", 0,  //$NON-NLS-1$
                         "", e )); //$NON-NLS-1$
        	}
            CoordinateReferenceSystem crsSys = null;
            if (prjRead != null) {
            	crsSys = prjRead.getCoodinateSystem();	
            }
            else {
            	//prj file not read, default to lat long
            	crsSys = DefaultGeographicCRS.WGS84;
            }
             
            DefaultParameterDescriptor crs = new DefaultParameterDescriptor("crs", //$NON-NLS-1$
                    CoordinateReferenceSystem.class, null, crsSys);
            DefaultParameterDescriptor env = WorldImageFormat.ENVELOPE;
            
            // Stolen from WorldImageFormat, as mInfo is not externally accesible           
            HashMap<String,String> info1 = new HashMap<String,String>();
            info1.put("name", "WorldImage");  //$NON-NLS-1$//$NON-NLS-2$
            info1.put("description", //$NON-NLS-1$
                "A raster file accompanied by a spatial data file"); //$NON-NLS-1$
            info1.put("vendor", "Geotools"); //$NON-NLS-1$ //$NON-NLS-2$
            info1.put("docURL", "http://www.geotools.org/WorldImageReader+formats"); //$NON-NLS-1$ //$NON-NLS-2$
            info1.put("version", "1.0"); //$NON-NLS-1$ //$NON-NLS-2$
            return new ParameterGroup(new DefaultParameterDescriptorGroup(
                    info1, new GeneralParameterDescriptor[] { crs, env }));
            
        } catch (MalformedURLException e) {
            CatalogPlugin.getDefault().getLog().log(
                    new org.eclipse.core.runtime.Status(IStatus.WARNING, 
                    "net.refractions.udig.catalog", 0,  //$NON-NLS-1$
                    Policy.bind("AbstractRasterGeoResource.MalformedURLException"), e )); //$NON-NLS-1$
            return super.getReadParameters();
        } catch (IOException e) {
            CatalogPlugin.getDefault().getLog().log(
                    new org.eclipse.core.runtime.Status(IStatus.WARNING, 
                    "net.refractions.udig.catalog", 0,  //$NON-NLS-1$
                    Policy.bind("AbstractRasterGeoResource.IOException"), e )); //$NON-NLS-1$
            return super.getReadParameters();
        } catch (FactoryException e) {
            CatalogPlugin.getDefault().getLog().log(
                    new org.eclipse.core.runtime.Status(IStatus.WARNING, 
                    "net.refractions.udig.catalog", 0,  //$NON-NLS-1$
                    Policy.bind("AbstractRasterGeoResource.FactoryException"), e )); //$NON-NLS-1$
            return super.getReadParameters();
        }
    }
   
    /**
     * Describes this Resource.
     * @author mleslie
     * @since 0.6.0
     */
    public class IGeoResourceWorldImageInfo extends IGeoResourceInfo {
        IGeoResourceWorldImageInfo() {
            this.keywords = new String[] {
                    "WorldImage", "world image", ".gif", ".jpg", ".jpeg",   //$NON-NLS-1$ //$NON-NLS-2$//$NON-NLS-3$//$NON-NLS-4$ //$NON-NLS-5$
                    ".tif", ".tiff", ".png"};   //$NON-NLS-1$//$NON-NLS-2$//$NON-NLS-3$
            this.title = getIdentifier().getFile();
            this.name = this.title;
            this.description = getIdentifier().toString();
            this.bounds = getBounds();
        }
        
        /*
         * @see net.refractions.udig.catalog.IGeoResourceInfo#getBounds()
         */
        public ReferencedEnvelope getBounds() {
            if(this.bounds == null) {
                Envelope env = null;
                try {
                    GridCoverage source = (GridCoverage)findResource();
                    org.opengis.spatialschema.geometry.Envelope ptBounds = 
                        source.getEnvelope();
                    env = new Envelope( ptBounds.getMinimum(0), ptBounds.getMaximum(0),
                            ptBounds.getMinimum(1), ptBounds.getMaximum(1));
                    
                    CoordinateReferenceSystem geomcrs = 
                        source.getCoordinateReferenceSystem();
                    
                    this.bounds = new ReferencedEnvelope(env, geomcrs);
                    /*
                    if(geomcrs != null) {
                        if(!geomcrs.equals(CRS.decode("EPSG:4269"))) { //$NON-NLS-1$
                            bounds = JTS.transform(bounds, CRS.decode("EPSG:4269")); //$NON-NLS-1$
                        }
                    } else {
                        System.err.println("CRS unknown for WorldImage"); //$NON-NLS-1$
                    }
                    */
                } catch (Exception e) {
                    CatalogPlugin.getDefault().getLog().log(
                            new org.eclipse.core.runtime.Status(IStatus.WARNING, 
                            "net.refractions.udig.catalog", 0,  //$NON-NLS-1$
                            Policy.bind("WorldImageGeoResource.error.layer.bounds"), e )); //$NON-NLS-1$ //$NON-NLS-2$
                }
            }
            return this.bounds;
        }
    }
}
