/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.catalog.rasterings;

import java.io.IOException;
import java.io.Serializable;
import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.refractions.udig.catalog.IGeoResource;
import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.IServiceInfo;

import org.eclipse.core.runtime.IProgressMonitor;
import org.geotools.data.coverage.grid.AbstractGridFormat;
import org.geotools.data.coverage.grid.GridFormatFactorySpi;
import org.opengis.coverage.grid.GridCoverageReader;

/**
 * Provides a handle to a raster service allowing the service to be lazily 
 * loaded.
 * <p>
 * This class provides functionality common to services attached to GridCoverage
 * based resources.
 * </p>
 * @author mleslie
 * @since 0.6.0
 */
public abstract class AbstractRasterService extends IService {
    private URL id;
    /** <code>status</code> field describes the status of the service */
    protected Status status = Status.NOTCONNECTED;
    /** <code>message</code> field reports any errors encountered.  May be null. */
    protected Exception message = null;
    /** <code>reader</code> field */
    protected GridCoverageReader reader;
    private AbstractGridFormat format;
    private GridFormatFactorySpi factory;

    /**
     * Construct <code>AbstractRasterService</code>.
     *
     * @param id
     * @param factory
     */
    public AbstractRasterService(URL id, GridFormatFactorySpi factory) {
        this.id = id;
        this.factory = factory;
    }

    public boolean canResolve(Class adaptee) {
        if(GridCoverageReader.class.isAssignableFrom(adaptee)) {
            return true;
        }
        if(IServiceInfo.class.isAssignableFrom(adaptee)) {
            return true;
        }
        if(IGeoResource.class.isAssignableFrom(adaptee)) {
            return true;
        }
        if(List.class.isAssignableFrom(adaptee)) {
            return true;
        }
        return false;
    }

    public Status getStatus() {
        return this.status;
    }
    
    public Throwable getMessage() {
        return this.message;
    }
    
    public URL getIdentifier() {
        return this.id;
    }

    /**
     * Finds or creates the GridFormat object describing the service.
     * 
     * @return GridFormat describing this coverage.
     */
    public Object getFormat() {
        if(this.format == null) {
            this.format = (AbstractGridFormat)this.factory.createFormat();
        }
        return this.format;
    }
    
    /**
     * Finds or creates the Reader used to access this service.  Apon any 
     * exception, the message field is populated and null is returned.
     * 
     * @return Reader linked to this service.
     */
    public GridCoverageReader getReader() {
        if(this.reader == null) {
            try {
                AbstractGridFormat frmt = (AbstractGridFormat)getFormat();
                this.reader = frmt.getReader(getIdentifier());
            } catch(Exception ex) {
                this.message = ex;
            }
        }
        return this.reader;
    }
    
    /**
     * Retrieves the string identifying this service.  This is the location
     * of the resource to be loaded.
     * 
     * @return String describing this service.
     */
    public String getDescription() {
        return getIdentifier().toString();
    }
    
    /**
     * Retrieves a relatively human readable title for this service.
     * 
     * @return Title of this service
     */
    public String getTitle() {
        return getIdentifier().getFile();
    }

    @Override
    public <T> T resolve( Class<T> adaptee, IProgressMonitor monitor ) 
            throws IOException {
        if(monitor != null)
            monitor.beginTask(Policy.bind("AbstractRasterService.resolve"), 3); //$NON-NLS-1$
        if(adaptee == null) {
            if(monitor != null)
                monitor.done();
            return null;
        }
        if(monitor != null)
            monitor.worked(1);
        if(adaptee.isAssignableFrom(GridCoverageReader.class)) {
            if(monitor != null)
                monitor.done();
            return adaptee.cast(getReader());
        }
        if(monitor != null)
            monitor.worked(1);
        if(adaptee.isAssignableFrom(IService.class)) {
            if(monitor != null)
                monitor.done();
            return adaptee.cast(this);
        }
        if(monitor != null)
            monitor.worked(1);
        if(adaptee.isAssignableFrom(IServiceInfo.class)) {
            if(monitor != null)
                monitor.done();
            return adaptee.cast(getInfo(monitor));
        }
        if(monitor != null)
            monitor.worked(1);
        if(adaptee.isAssignableFrom(List.class)) {
            if(monitor != null)
                monitor.done();
            return adaptee.cast( members(monitor));
        }
        if(monitor != null)
            monitor.done();
        return null;
    }

    @Override
    public abstract List<AbstractRasterGeoResource> members( IProgressMonitor monitor ) 
            throws IOException;

    @Override
    public Map<String, Serializable> getConnectionParams() {
        return new HashMap<String, Serializable>();
    }
    
    public abstract IServiceInfo getInfo(IProgressMonitor monitor) throws IOException;
}
