package net.refractions.udig.catalog.ui;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.refractions.udig.catalog.CatalogPlugin;
import net.refractions.udig.catalog.internal.ui.ConnectionPage;
import net.refractions.udig.catalog.internal.ui.Policy;
import net.refractions.udig.catalog.internal.ui.URLWorkflowAction;
import net.refractions.udig.core.internal.ExtensionPointProcessor;
import net.refractions.udig.core.internal.ExtensionPointUtil;
import net.refractions.udig.internal.ui.UiPlugin;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.jface.dialogs.IDialogSettings;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.ImageRegistry;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.wizard.IWizardPage;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.jface.wizard.WizardSelectionPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;

/**
 * This is a wizard page which is used to select from a set of data sources.
 * <p>
 * Each data source is described an instance of
 * 
 * @see net.refractions.udig.catalog.ui.UDIGImportPageDescriptor.
 *      </p>
 *      <p>
 *      Client code may instantiate objects of this class,but may only call
 *      setSelectedDescriptor(UDIGImportPageDescriptor). All other methods are
 *      wizard lifecycle methods.
 *      </p>
 * @author Justin Deoliveira,Refractions Research Inc.,jdeolive@refractions.net
 */
public class DataSourceSelectionPage extends WizardSelectionPage implements
		ISelectionChangedListener {

	/** the last selected datasource **/
	static final String LASTDS_KEY 
		= "DataSourceSelectionPage.lastDataSourceSelected"; //$NON-NLS-1$
	
	/** the selected extension * */
	private IStructuredSelection selected;

	/** list of wizard page extension* */
	private List<UDIGImportPageDescriptor> descriptors;

	/** the viewer to select * */
	private WizardViewer viewer;
    
    /**
     * Used to cache the pages, rather than create them every time getNextPage()
     * is called. This lets us maintain data that the user has entered if they
     * backtrack.
     */
    private Map<UDIGImportPageDescriptor, UDIGImportPage> pageCache = 
        new HashMap<UDIGImportPageDescriptor, UDIGImportPage>();

	public DataSourceSelectionPage() {
		super(Policy.bind("CatalogImportWizard.pageName"));//$NON-NLS-1$

		setMessage(Policy.bind("CatalogImportWizard.pageMessage")); //$NON-NLS-1$
	}

	/**
	 * Sets the selected set of import page descriptiors. 
	 * 
	 * @param ids A list of import page identifiers.
	 * 
	 */
	public void select(List<String> ids) {
		//process extension point, filter by id
		UDIGImportPageProcessor p = new UDIGImportPageProcessor(ids);
		String xpid = UDIGImportPage.XPID;
		ExtensionPointUtil.process(CatalogUIPlugin.getDefault(), xpid, p);
		
		setSelection(p.descriptors);
	}

	
	
	/**
	 * Sets the selected import page descriptor.
	 * 
	 * @param id The id of the import page to select.
	 */
	public void select(String id) {
		ArrayList<String> l = new ArrayList<String>();
        l.add(id);
        select(l);
	}
	
	/**
	 * Sets the selected import page descriptor.
	 * 
	 * @param descriptor The descriptor to be selected.
	 */
	public void setSelection(List<UDIGImportPageDescriptor> descriptors) {
		selected = new StructuredSelection(descriptors);
		if (viewer != null) {
			//only set first in viewer
			if (!selected.isEmpty()) {
				viewer.setSelection(
					new StructuredSelection(selected.getFirstElement())
				);
			}
			else {
				viewer.setSelection(new StructuredSelection());
			}
		}
	}
	
	@Override
	public boolean canFlipToNextPage() {
		return isPageComplete();
	}

	@Override
	public boolean isPageComplete() {
		//if selection set, we are done
		if (selected != null && selected.size() == 1) {
			return true;
		}
			
		//if there is only one choice, we are also done
		if (descriptors != null && descriptors.size() == 1)
			
			return true;
			
		return false;
	}

	@Override
	public IWizardPage getNextPage() {
		UDIGImportPageDescriptor d 
			= (UDIGImportPageDescriptor) selected.getFirstElement();
		if (d == null) {
			if (descriptors != null && descriptors.size() == 1)
				d = descriptors.get(0);
		}
		
		if (d != null) {
			try {
				IDataWizard wizard = (IDataWizard) getWizard();

				UDIGImportPage page = pageCache.get(d); 
                if (page == null) {
                    page = d.createImportPage();
                    pageCache.put(d, page);
                }
				wizard.init((WizardPage) page);

				ConnectionPage next = new ConnectionPage(page);
				wizard.init(next);

				return next;
			} 
			catch (CoreException e) {
				String msg = "Could not instantiate import wizard page"; //$NON-NLS-1$
				CatalogPlugin.log(msg, e);
			}
		}

		return null;
	}

	/**
	 * @see org.eclipse.jface.dialogs.IDialogPage#createControl(org.eclipse.swt.widgets.Composite)
	 */
	public void createControl(Composite parent) {
		
		viewer = new WizardViewer(parent, SWT.SINGLE);
		viewer.setInput(getDescriptors().toArray());

		viewer.addDoubleClickListener(new IDoubleClickListener() {
			public void doubleClick(DoubleClickEvent event) {
				IWizardPage next = getNextPage();
				getWizard().getContainer().showPage(next);
			}
		});
		
		//check dialog setting for last selection
		IDialogSettings settings = getDialogSettings();
		String last = settings.get(LASTDS_KEY);
		if (last != null) select(last);
		
		viewer.addSelectionChangedListener(this);
		setControl(viewer.getControl());
	}

	/**
	 * @see org.eclipse.jface.viewers.ISelectionChangedListener#selectionChanged(org.eclipse.jface.viewers.SelectionChangedEvent)
	 */
	public void selectionChanged(SelectionChangedEvent event) {
		selected = (IStructuredSelection) event.getSelection();
		if (selected == null || selected.isEmpty())
			return;
		
		UDIGImportPageDescriptor descriptor 
			= (UDIGImportPageDescriptor) selected.getFirstElement();
		
		setDescription(descriptor.getDescription());
		setImageDescriptor(descriptor.getDescriptionImage());

		getWizard().getContainer().updateButtons();
	}

	protected List<UDIGImportPageDescriptor> getDescriptors() {

		if (descriptors == null) {
			UDIGImportPageProcessor p = new UDIGImportPageProcessor();
			String xpid = UDIGImportPage.XPID;

			ExtensionPointUtil.process(CatalogUIPlugin.getDefault(), xpid, p);
			descriptors = p.descriptors;
		}
        Collections.sort(descriptors, new Comparator() {
            public int compare(Object o1, Object o2) {
                String s1 = ((UDIGImportPageDescriptor)o1).getLabel();
                String s2 = ((UDIGImportPageDescriptor)o2).getLabel();
                return s1.compareTo(s2);
            }
        });
		return descriptors;
	}
	
	protected void saveSelectedDescriptor() {
		IStructuredSelection selection = 
			(IStructuredSelection) viewer.getSelection();
		if (selection.isEmpty())
			return;
		
		UDIGImportPageDescriptor selected = 
			(UDIGImportPageDescriptor) selection.getFirstElement();
		
		IDialogSettings settings = getDialogSettings();
		settings.put(LASTDS_KEY, selected.getId());
	}

	private class UDIGImportPageProcessor implements ExtensionPointProcessor {

		List<UDIGImportPageDescriptor> descriptors = new ArrayList<UDIGImportPageDescriptor>();
		List<String> ids;
		
		UDIGImportPageProcessor() {
			this(null);
		}
		
		UDIGImportPageProcessor(List<String> ids) {
			this.ids = ids;
		}

		public void process(IExtension extension, IConfigurationElement element)
				throws Exception {

			UDIGImportPageDescriptor descriptor = 
				new UDIGImportPageDescriptor(element);
			
			if (ids != null) {
				if (ids.contains(descriptor.getId())) {
					descriptors.add(descriptor);
				}
			}
			else descriptors.add(descriptor);
		}
	}

	private class WizardViewer extends TableViewer {

		public WizardViewer(Composite parent, int style) {
			super(parent, style);

			setContentProvider(new ArrayContentProvider());
			setLabelProvider(new LabelProvider() {
				public String getText(Object object) {
					UDIGImportPageDescriptor descriptor = (UDIGImportPageDescriptor) object;

					return descriptor.getLabel();
				}

				public Image getImage(Object object) {
					UDIGImportPageDescriptor descriptor = (UDIGImportPageDescriptor) object;

					String id = descriptor.getId();
					ImageRegistry registry = UiPlugin.getDefault()
							.getImageRegistry();
					ImageDescriptor image = descriptor.getImage();

					if (registry.get(id) == null && image != null) {
						registry.put(id, image);
					}

					return registry.get(id);
				}
			});
		}
	}
}
