/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.catalog.ui;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.refractions.udig.catalog.CatalogPlugin;
import net.refractions.udig.catalog.ICatalog;
import net.refractions.udig.catalog.IGeoResource;
import net.refractions.udig.catalog.IResolve;
import net.refractions.udig.catalog.IResolveChangeEvent;
import net.refractions.udig.catalog.IResolveChangeListener;
import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.internal.ui.Policy;

import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.widgets.Display;

/**
 * Provides a threaded Tree content provider for IResolve.
 * <p>
 * Responsibilities:
 * <ul>
 * <li>Rooted by a catalog
 * <li>Ensure that calls to members are dispatched in a non ui thread
 * </ul>
 * </p>
 * @author jgarnett
 * @since 0.6.0
 */
public class ResolveContentProvider implements ITreeContentProvider, IResolveChangeListener {
    /**
     * Map of threads used to resolve associated IResolves.
     * <p>
     * Waking up the thread will cause the associated IResolve to be refreshed.
     * </p>
     */
    Map<IResolve,Thread> threadFarm = new HashMap<IResolve,Thread>();
    
    /**
     * Captures parent child relationships.
     */
    Map<IResolve,List<IResolve>> structure = new HashMap<IResolve,List<IResolve>>();

    /**
     * Root of this tree, often a ICatalog.
     */
    private ICatalog catalog;
    private List<IResolve> list;

    private Viewer viewer;
    
    /**
     * Catalog has changed!
     * <p>
     * Will only start up a thread to update the structure if:
     * <ul>
     * <li>We care about this resolve
     * <li>We are not already running a thread
     * </ul>
     * </p>
     * This will allow us to "ignore" events generated by the process
     * up inspecting the resolve being updated.
     * </p> 
     * @param event
     */
    public void changed( IResolveChangeEvent event ) {
        if( event.getType() != IResolveChangeEvent.Type.POST_CHANGE ) {
            return;
        }
        IResolve resolve = event.getResolve();
        if( threadFarm.containsKey( resolve )){
            update( resolve );
        }
    }

  
    /**
     * Called by thread to client code to get the content refreshed.
     * <p>
     * This call will not update structure, just appearance.
     * </p>
     * @param resolve
     */
    public void refresh( final IResolve resolve ){
        Display.getDefault().asyncExec(new Runnable() {
            public void run() {
                if( viewer instanceof TreeViewer){
                    ((TreeViewer)viewer).refresh(resolve);
                }
                else {
                    viewer.refresh();
                }
            }
        });
    }
    /**
     * Update appearance and structure.
     * <p>
     * Note: this will spawn a thread to fetch the required information.
     * </p>
     * @param resolve
     */
    public void update( final IResolve resolve ){
        if( resolve == null ){
            // go away
            return;
        }
        if( resolve.getIdentifier() == null ){
            // go away x 2
            //System.out.println( "Got an resolve with out an id "+ resolve);            
        }
        // run the thread, unless it is already running...
        // (this will cause any change events generated by the thread
        // to be ignored).
        if( threadFarm.containsKey( resolve ) ){ 
            Thread update = threadFarm.get( resolve );
            if( update == null ){
                // This IResolve is not suitable for member resolution
                //
            }
            if( update.isAlive() ){
                // thread will already report back to structure            
                // Note: thread should end with a async update to the
                // assoicated element                
            }
            else {
                // We had a thread but it stopped - must be do to an error?
                //
                update.run();
            }
        }
        else {
            // request a structure update
            Thread update = new Thread( new Update( resolve ), "Update "+resolve.getIdentifier() ); //$NON-NLS-1$
            threadFarm.put( resolve, update );
            update.setPriority( Thread.MIN_PRIORITY );
            update.run();
        }
    }
    
    /**
     * Thread for updating structure
     * <p>
     * Note: thread notify the system that the element has requires an update
     */
    class Update implements Runnable{
        IResolve resolve;
        Update( IResolve target ){
            resolve = target;
        }
        /**
         * Update strucuture, Thread will be notified if more updates are required.
         * <p>
         * Note: We also need to let ourselves be interrupted
         */
        public void run() {
            try {                
                //System.out.println( resolve.getIdentifier()+": update requested" );
                try {
                    //System.out.println( resolve.getIdentifier()+": request memebers " );
                    List members = resolve.members( null );
                    if( members != null && members.size() != 0 ){
                        //System.out.println( resolve.getIdentifier()+": found "+members.size()+" memebers" );    
                    }                    
                    structure.put( resolve, members );                     
                }
                catch( IOException io ){
                    // could not get children
                    //System.out.println( resolve.getIdentifier()+": "+ io );
                    io.printStackTrace();
                    structure.put( resolve, null );
                    //return; // don't even try again
                }
                refresh( resolve );
            } 
            catch (Throwable t ){
                //System.out.println( resolve.getIdentifier()+": "+t );
            }
        }
    }
    /**
     * Returns the child elements of the given parent element.
     * <p>
     * The difference between this method and <code>IStructuredContentProvider.getElements</code> 
     * is that <code>getElements</code> is called to obtain the 
     * tree viewer's root elements, whereas <code>getChildren</code> is used
     * to obtain the children of a given parent element in the tree (including a root).
     * </p>
     * The result is not modified by the viewer.
     * </p>
     * @see org.eclipse.jface.viewers.ITreeContentProvider#getChildren(java.lang.Object)
     *
     * @param parent the parent element
     * @return an array of child elements
     * 
     */
    public Object[] getChildren(Object parent) {
        if( parent == null ) return null;
        if( parent instanceof List) return ((List)parent).toArray();
        if( parent instanceof String ) return new Object[0];
        if( !(parent instanceof IResolve) ) return null;
        
        IResolve resolve = (IResolve) parent;
        if( structure.containsKey( resolve )){
            List<IResolve> members = structure.get( resolve );
            return members != null ? members.toArray() : null;
        }
        else {
            update( resolve ); // calculate            
            return new Object[]{ Policy.bind("ResolveContentProvider.searching", CatalogUIPlugin.label( resolve ))};             //$NON-NLS-1$
        }
    }
    
    /**
     * @see org.eclipse.jface.viewers.ITreeContentProvider#getParent(java.lang.Object)
     */
    public Object getParent(Object element) {
        if( !(element instanceof IResolve)) return null;
        IResolve resolve = (IResolve) element;               
        try{
            // assume find parent is cheap        
            if( element instanceof IService ) {
                if( list!=null && list.contains( resolve ) ) return list;
                if( catalog != null ) return catalog;
                return null; // service is probably the parent                
            }
            if( element instanceof IGeoResource ) {
                return ((IGeoResource) element).resolve(IService.class,null);                
            }
            if( element instanceof String ) {
                return null;
            }
        } catch(Throwable t){
            t.printStackTrace();
        }
        return null;
    } 
    
    /**
     * @see org.eclipse.jface.viewers.ITreeContentProvider#hasChildren(java.lang.Object)
     */
    public boolean hasChildren(Object element) {
        try{
        if(element == null)
            return false;
        if(element instanceof ICatalog)
    //        try{
    //            Iterator i = ((ICatalog)element).services();
    //            return i!=null && i.hasNext();
    //        }catch(Throwable t){
    //            t.printStackTrace();
    //            return false;
    //        }
            return true;
        if(element instanceof IService)
    //        try{
    //            Iterator i = ((IService)element).resources();
    //            return i!=null && i.hasNext();
    //        }catch(Throwable t){
    //            t.printStackTrace();
    //            return false;
    //        }
            return true;
        if(element instanceof IResolve){            
            return true;
        }
        return ((element instanceof List) && !((List)element).isEmpty());
        }catch(Throwable t){
            t.printStackTrace();
        }
        return false;
    }
    
    /**
     * @see org.eclipse.jface.viewers.IStructuredContentProvider#getElements(java.lang.Object)
     */
    public Object[] getElements(Object inputElement) {
        return getChildren(inputElement);
    }

    /**
     * Notifies this content provider that the given viewer's input
     * has been switched to a different element.
     * <p>
     * A typical use for this method is registering the content provider as a listener
     * to changes on the new input (using model-specific means), and deregistering the viewer 
     * from the old input. In response to these change notifications, the content provider
     * should update the viewer (see the add, remove, update and refresh methods on the viewers).
     * </p>
     * <p>
     * The viewer should not be updated during this call, as it might be in the process
     * of being disposed.
     * </p>
     *
     * @param viewer the viewer
     * @param oldInput the old input element, or <code>null</code> if the viewer
     *   did not previously have an input
     * @param newInput the new input element, or <code>null</code> if the viewer
     *   does not have an input
     */
    public void inputChanged( Viewer newViewer, Object oldInput, Object newInput ) {        
        if( oldInput == newInput ){
            return;
        }
        viewer = newViewer;
        if( catalog != null || list != null ){
            CatalogPlugin.removeListener( this );
        }
        catalog = newInput instanceof ICatalog ? (ICatalog) newInput : null;
        list = newInput instanceof List ? (List<IResolve>) newInput : null;
        if( catalog != null || list != null ){
            CatalogPlugin.addListener( this );
        }
    }
    /*
     * @see org.eclipse.jface.viewers.IContentProvider#dispose()
     */
    public void dispose() {
        if( threadFarm != null ){
            for( IResolve resolve : threadFarm.keySet() ){
                Thread thread = threadFarm.get( resolve );
                if( thread.isAlive() ){
                    thread.interrupt();                    
                }                
            }
            threadFarm.clear();
            threadFarm = null;
        }
        if( structure != null ){
            for( IResolve resolve : structure.keySet() ){
                List<IResolve> children = structure.get( resolve );
                children.clear();                                
            }
            structure.clear();
        }
    }
}