/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.catalog.ui;

import java.io.IOException;
import java.util.List;

import net.refractions.udig.catalog.CatalogPlugin;
import net.refractions.udig.catalog.ICatalog;
import net.refractions.udig.catalog.IResolve;
import net.refractions.udig.catalog.IResolveChangeEvent;
import net.refractions.udig.catalog.IResolveChangeListener;
import net.refractions.udig.catalog.IService;

import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.widgets.Display;

/**
 * Provides a threaded Tree content provider for IResolve.
 * <p>
 * Responsibilities:
 * <ul>
 * <li>Rooted by a catalog
 * <li>Ensure that calls to members are dispatched in a non ui thread
 * </ul>
 * </p>
 * @author jgarnett
 * @since 0.6.0
 */
public class ResolveContentProviderSimple implements ITreeContentProvider, IResolveChangeListener {
    /** May be catalog, list or string */
    Object content;
    /** Last viewer to use us */
    Viewer viewer;
    /**
     * Catalog has changed!
     * @param event
     */
    public void changed( IResolveChangeEvent event ) {
        if( event.getType() != IResolveChangeEvent.Type.POST_CHANGE ) {
            return;
        }
        IResolve resolve = event.getResolve();
        refresh( resolve ); // may not be one of our elements but who cares        
    }

  
    /**
     * Called by thread to client code to get the content refreshed.
     * <p>
     * This call will not update structure, just appearance.
     * </p>
     * @param resolve
     */
    public void refresh( final IResolve resolve ){
        Display.getDefault().asyncExec(new Runnable() {
            public void run() {
                if( viewer instanceof TreeViewer){
                    ((TreeViewer)viewer).refresh(resolve);
                }
                else {
                    viewer.refresh();
                }
            }
        });
    }
    
    /**
     * Returns the child elements of the given parent element.
     * <p>
     * The difference between this method and <code>IStructuredContentProvider.getElements</code> 
     * is that <code>getElements</code> is called to obtain the 
     * tree viewer's root elements, whereas <code>getChildren</code> is used
     * to obtain the children of a given parent element in the tree (including a root).
     * </p>
     * The result is not modified by the viewer.
     * </p>
     * @see org.eclipse.jface.viewers.ITreeContentProvider#getChildren(java.lang.Object)
     *
     * @param parent the parent element
     * @return an array of child elements
     * 
     */
    public Object[] getChildren(Object parent) {
        if( parent == null ) return null;
        if( parent instanceof List) return ((List)parent).toArray();
        if( parent instanceof String ) return new Object[0];
        if( !(parent instanceof IResolve) ) return null;
        
        IResolve resolve = (IResolve) parent;
        try {
            return resolve.members( null ).toArray();
        } catch (IOException e) {
            return new Object[]{ e.getLocalizedMessage() };
        }        
    }
    
    /**
     * @see org.eclipse.jface.viewers.ITreeContentProvider#getParent(java.lang.Object)
     */
    public Object getParent(Object element) {
        if( !(element instanceof IResolve)) return null;
        IResolve resolve = (IResolve) element;
        
        // assume find parent is cheap
        if( content instanceof List ) {
            List list = (List) content;
            if( list.contains( element ) ){
                return list;
            }
            else {
                return null; // we are a search and we don't know our parent
            }
        }
        try {
            return resolve.parent( null );
        } catch (IOException e) {
            return null;
        }         
    } 
    
    /**
     * @see org.eclipse.jface.viewers.ITreeContentProvider#hasChildren(java.lang.Object)
     */
    public boolean hasChildren(Object element) {
        if(element == null) return false;
        if(element instanceof List ) return true;
        if(element instanceof ICatalog) return true;
        if(element instanceof IService) return true;
        if(element instanceof IResolve) {
            IResolve resolve = (IResolve) element;
            try {
                return resolve.members( null ) != null;
            } catch (IOException e) {
                return false;
            }
        }
        return false;
    }
    
    /**
     * @see org.eclipse.jface.viewers.IStructuredContentProvider#getElements(java.lang.Object)
     */
    public Object[] getElements(Object inputElement) {
        return getChildren(inputElement);
    }

    /**
     * Notifies this content provider that the given viewer's input
     * has been switched to a different element.
     * <p>
     * A typical use for this method is registering the content provider as a listener
     * to changes on the new input (using model-specific means), and deregistering the viewer 
     * from the old input. In response to these change notifications, the content provider
     * should update the viewer (see the add, remove, update and refresh methods on the viewers).
     * </p>
     * <p>
     * The viewer should not be updated during this call, as it might be in the process
     * of being disposed.
     * </p>
     *
     * @param viewer the viewer
     * @param oldInput the old input element, or <code>null</code> if the viewer
     *   did not previously have an input
     * @param newInput the new input element, or <code>null</code> if the viewer
     *   does not have an input
     */
    public void inputChanged( Viewer newViewer, Object oldInput, Object newInput ) {        
        viewer = newViewer;
        if( content == newInput ){
            return;
        }        
        if( content != null ){
            if( content instanceof ICatalog ){
                ICatalog catalog = (ICatalog) content;
                catalog.removeCatalogListener( this );
            }
            else {
                CatalogPlugin.removeListener( this );
            }
        }
        content = newInput;
        if( content != null ){
            if( content instanceof ICatalog ){
                ICatalog catalog = (ICatalog) content;
                catalog.addCatalogListener( this );
            }
            else {
                CatalogPlugin.addListener( this );
            }
        }
    }
    
    /*
     * @see org.eclipse.jface.viewers.IContentProvider#dispose()
     */
    public void dispose() {
        if( content != null ){
            if( content instanceof ICatalog ){
                ICatalog catalog = (ICatalog) content;
                catalog.addCatalogListener( this );
            }
            else {
                CatalogPlugin.addListener( this );
            }
        }
    }
}