/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.catalog.ui;

import java.io.IOException;
import java.net.URL;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.NoSuchElementException;

import net.refractions.udig.catalog.CatalogPlugin;
import net.refractions.udig.catalog.IGeoResource;
import net.refractions.udig.catalog.IResolve;
import net.refractions.udig.catalog.IResolveChangeEvent;
import net.refractions.udig.catalog.IResolveChangeListener;
import net.refractions.udig.catalog.internal.Policy;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;

/**
 * Label generation for IResolve (must be non-blocking and quick).
 * <p>
 * Compare and contrast with ResovleLabelDecorator which is allowed to block.
 * </p>
 * 
 * @author jgarnett
 * @since 0.7.0
 */
public class ResolveLabelProvider extends LabelProvider {
    
    /**
     * ImageCache for images *we* are responsible for.
     * <p>
     * Keys are IResolve.getID().toExternalForm(), Icons go in, images come out.
     */
    //ImageRegistry imageCache;
    
    /**
     * LabelCache for titles *we* have requested.
     * <p>
     * Keys are IResolve.getID().toExternalForm(), Icons go in, images come out.
     */   
    Map<String,String> labelCache = Collections.synchronizedMap( new HashMap<String,String>() );
    
    /**
     * Place keys in here w/ False when we put them in the queue, change to True when executed. 
     */
    Map<String,Boolean> requestStatus = Collections.synchronizedMap( new HashMap<String,Boolean>() );
    
    /**
     * Queue of IResolves needing to be refreshed.
     * <p>
     * Does not allow duplicates to be added.
     * </p>
     */
    LinkedList<IResolve> queue = new LinkedList<IResolve>(){
        /** <code>serialVersionUID</code> field */
        private static final long serialVersionUID = 3258410638249112883L;
        public void add(int index,IResolve resource){
            if( !contains( resource ) )
                super.add( index, resource );            
        }
        public boolean add(IResolve resource){
            if( !contains( resource ) )
                return super.add( resource );
            return false;
        }
        public void addFirst(IResolve resource){
            if( !contains( resource) )
                super.addFirst( resource );
                
        }
        public void addLast(IResolve resource){
            if( !contains( resource) )
                super.addFirst( resource );
        }
    };
    
    /** Will only send on refresh events for IResolves we know and care about */
    IResolveChangeListener watcher = new IResolveChangeListener(){
        public void changed( IResolveChangeEvent event ) {
            try {
                if( event.getType() != IResolveChangeEvent.Type.POST_CHANGE 
                		&& event.getType() != IResolveChangeEvent.Type.PRE_DELETE) return;
                
                IResolve resolve = event.getResolve();
                // TODO: look through delta to find parents
                if( resolve == null ) return;
                
                URL identifier = resolve.getIdentifier();
                if( identifier == null ) return;
                
                //String KEY = identifier.toExternalForm();
                //System.out.println( KEY +": catalog event "+event );
                
                /*
                            
                if( queue.contains( event.getResolve() )){
                    //System.out.println( KEY +": catalog event ignored as we are enqueued" );
                    return; // we are already going to work on this one
                }
                if( requestStatus.containsKey( KEY )){
                    if( requestStatus.get( KEY )){ // we got one
                        requestStatus.remove( KEY );
                        // force it to be requested again
                        // (note image and title are still there until replaced)
                        // refresh( resolve );
                        return;
                    }
                    else {
                        //System.out.println( KEY +": catalog event ignored - we know it lacks title/icon" );
                        return;
                    }
                }*/
            }
            catch( Throwable t ){
                CatalogUIPlugin.getDefault().getLog().log( new Status(IStatus.ERROR, CatalogUIPlugin.ID, IStatus.OK, null, t));
            }
        }        
    };
    /** 
     * Piccaso generates pcitures for layers in the queue.
     * <p>
     * This is the sole provider of dynamic artwork for layers.
     * Piccaso will block contacting external servers and so on.
     * </p>
     * <p>
     * If this gets to be a pain we may switch to the dutch school
     * model, perhaps even impressionests based on a sample feature.
     * </p>
     * <p>
     * Any artwork is provided as an ImageDescriptor using the key
     * GENERATED_ICON. This will be turned into an Image by the
     * decorateImage method as required.
     * </p>
     */
    Runnable picasso = new Runnable(){public void run() {
        IResolve resource = null;
        RESOLVE: while( queue!= null ) {
            synchronized (queue) {                
                try {
                    while( queue != null && queue.isEmpty() )
                        queue.wait(1000); // yield thread and wait for something to do
                } catch (InterruptedException e) {
                    break RESOLVE; // exit runable cause we are interrupted
                }
                try {
                    if( queue == null ) break RESOLVE;
                    //System.out.println("Queue("+queue.size()+") removing ... ");
                    resource = queue.removeFirst();
                    //System.out.println("Queue("+queue.size()+") removed "+resource);
                } catch (NoSuchElementException notYet) {
                    continue RESOLVE;
                }
            }
            try {
                boolean refresh = false;
                if (resource == null || resource.getStatus() == IResolve.Status.BROKEN ){
                    continue RESOLVE;
                }                    
                final String KEY = resource.getIdentifier().toExternalForm();
                requestStatus.put( KEY, true );
                
                //System.out.println( KEY+": processing ..." );
                /*
                try{ // grab icon                    
                    ImageDescriptor icon = CatalogUIPlugin.icon( resource );
                    if( icon != null && imageCache.getDescriptor( KEY ) != icon ){
                        imageCache.put( KEY, icon );
                        //System.out.println( KEY+": fetching image for " + icon );                        
                        imageCache.get( KEY );                        
                        //System.out.println( KEY+": image is " + icon );                        
                        refresh = true;
                    }
                }      
                catch (IOException io ){
                    CatalogPlugin.log( KEY, io );
                }                    
                catch( Throwable t ){
                    CatalogPlugin.log( KEY, t );
                }
                */
                try { // grab title
                    
                    String title = CatalogUIPlugin.title( resource );
                    if( title != null && !title.equals( labelCache.get(KEY ))){
                        //System.out.println( KEY+": title is " + title );                        
                        labelCache.put( KEY, title );
                        refresh = true;
                    }
                }
                catch (IOException io ){
                    io.printStackTrace();
                }                    
                catch( Throwable t ){
                    // ignore
                }
                
                if( refresh ) {
                    refresh( resource );                    
                }
                else {
                    //System.out.println( KEY+": no title or icon available" );
                }
            }
            catch (Throwable t ){
                // must catch all icon errors or this thread will die :-P
            }
        }
    }};
    
    /**
     * Ask the ui thread to refresh our appearance.
     * <p>
     * Called for:
     * <ul>
     * <li>Catalog change events via watcher
     * <li>picasso when it has new stuff in imageCache or titleCache for us
     * </ul>
     */
    void refresh( final IResolve resource ) {      
        String KEY = resource.getIdentifier().toExternalForm();
        
        // need to update in the event thread
        //
        /*
        Display.getDefault().asyncExec( new Runnable(){
            public void run() {
                fireLabelProviderChanged(
                        new LabelProviderChangedEvent(ResolveLabelProvider.this, resource)
                );
            }
        });
        */
    }
    public ResolveLabelProvider(){
        this( Display.getDefault() );        
    }    
    public ResolveLabelProvider( Display display ){
        // imageCache = new ImageRegistry( display );
        Thread t = new Thread(picasso);
        t.start();
        t.setPriority( Thread.MIN_PRIORITY );
        CatalogPlugin.addListener( watcher );
    }
    
    /**
     * Generate text from the resolve.getURI()
     * <p>
     * Note this name is only used as a first try, the ResolveLabelDecorator
     * is expected to provide a label based on Name or Title information.
     * </p>
     * @param element
     * @return label based on IResolve.getIdentifier
     */
    public String getText( Object element ) {
        if( element == null || !(element instanceof IResolve)){
            return Policy.bind("ResolveLabelProvider.missingText"); //$NON-NLS-1$
        }        
        if( !(element instanceof IResolve) ){
            return super.getText( element );
        }
        IResolve resource = (IResolve) element;        
        String KEY = resource.getIdentifier().toExternalForm();
        
        if( !(resource instanceof IGeoResource) ){
            return CatalogUIPlugin.label( resource );
        }
        
        if( !requestStatus.containsKey( KEY ) ){
            queue( resource ); // we don't know about this one yet
        }        
        if( labelCache.containsKey( KEY )){ 
            String title = labelCache.get( KEY );
            if( title != null ){
                //System.out.println( KEY+":real title "+title+ " returned");
                return title;
            }
        }
        return CatalogUIPlugin.label( resource );
    }

    /**
     * Obtain image for the provided element.
     * <p>
     * To accomplish this quickly we simply make use of constants from
     * CatalogPlugin. We need a second pass that makes use of the real icon
     * from the real resource.
     * </p>
     * @param element is expeced to be IResolve
     * @return the image used to label the element, or <code>null</code>
     *   if there is no image for the given object
     * @see org.eclipse.jface.viewers.LabelProvider#getImage(java.lang.Object)
     */
    public Image getImage( Object element ) {
        if( element == null ){
            return ImageDescriptor.getMissingImageDescriptor().createImage();
        }
        if( !(element instanceof IResolve) ){
            return super.getImage( element );
        }
        IResolve resource = (IResolve) element;
        /*
        String KEY = resource.getIdentifier().toExternalForm();
        
        if( !requestStatus.containsKey( KEY ) ){
            queue( resource ); // we don't know about this one yet
        }
        
        if( imageCache.getDescriptor( KEY ) != null ){                
            Image image = imageCache.get( KEY );
            if( image.getBounds().width == 16 && image.getBounds().height == 16 ){
                return image;
            }
            //System.out.println( KEY+": image was not 16x16"); //$NON-NLS-1$
        }*/            
        return CatalogUIPlugin.image( resource );
    }    
    /** Enqueye resource, and wake up thread to make icon/title for resouce */
    private void queue( IResolve resource ){
        synchronized (queue) {
            String KEY = resource.getIdentifier().toExternalForm();
            
            if( requestStatus.containsKey( KEY )) return; // already requested or procesed
            
            requestStatus.put( KEY, Boolean.FALSE ); // mark requested            
            if( queue.add(resource) ) {
                //System.out.println( KEY+": queue added " );
                synchronized (queue) {
                    queue.notifyAll();
                }
            }
            else {
                //System.out.println( KEY+": queue already member" );
            }
        }
    }
    /*
     * @see org.eclipse.jface.viewers.LabelProvider#dispose()
     */
    public void dispose() {
        if( watcher != null ){
            CatalogPlugin.removeListener( watcher );
            watcher = null;
        }
        LinkedList hold = queue;
        if( hold != null ){
            queue = null;
            hold.clear();
            // hold.notifyAll(); 
            hold = null;
        }
        if( labelCache != null ){
            labelCache.clear();
            labelCache = null;
        }
        if( requestStatus != null ){
            requestStatus.clear();
            requestStatus = null;
        }
        
        super.dispose();
    }
}