package net.refractions.udig.internal.ui;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import net.refractions.udig.ui.IDropAction;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.DropTargetAdapter;
import org.eclipse.swt.dnd.DropTargetEvent;
import org.eclipse.swt.dnd.TransferData;

/**
 * A drop event handler which delegates to an extension point in order to 
 * perform drag and drop actions.
 * <p>
 * The drop handler interacts with three objects.
 * <ol>
 * 	<li>The <b>type</b> object, which is the entity being dropped.
 * 	<li>The <b>destination</b> object, which is where the drop is occuring. This
 * is usually a view or an editor.
 * 	<li>The <b>target</b> object (optional), which is the specific object within 
 * the destination.
 * </ol>
 * </p>
 * <p>
 * A UDIGDropHandler is delegated to by another DropTargetAdapter. Depending 
 * on the type of drop adapter, location, and target may not be set. 
 * </p>
 * 
 * @author Justin Deoliveira,Refractions Research Inc.,jdeolive@refractions.net
 * @since 0.6.0
 */
public class UDIGDropHandler extends DropTargetAdapter {
	
	/** job queue shared by all drop handlers **/
	private static CompositeDropActionJob jobQueue;
	
	/** the type object **/
	Object type;
	
	/** the destination object **/
	Object destination;
	
	/** the target object **/
	Object target;
	
	/** the location of teh drop **/
	int location = -1;
	
	/** procssed actions **/
	List<IDropAction> actions;
	
	/** adapter that created the handler **/
	DropTargetAdapter adapter;

	private TransferData[] transferType;
	
	public UDIGDropHandler(DropTargetAdapter adapter) {
		this.adapter = adapter;
	}
	
	public void setDestination(Object destination) {
		this.destination = destination;
	}
	
	public Object getDestination() {
		return destination;
	}
	
	public void setTarget(Object target) {
		this.target = target;
	}
	
	public Object getTarget() {
		return target;
	}
	
	public void setLocation(int location) {
		this.location = location;
	}
	
	public int getLocation() {
		return location;
	}
    
    
	
	@Override
    public void dragLeave( DropTargetEvent event ) {
        super.dragLeave(event);
    }

    @Override
    public void dragOver( DropTargetEvent event ) {
        event.detail = DND.DROP_COPY;
        
        if ((event.operations & DND.DROP_LINK) == DND.DROP_LINK) {
            event.detail = DND.DROP_LINK;
        }
    }

    @Override
    public void dropAccept( DropTargetEvent event ) {
        super.dropAccept(event);
    }

    @Override
    public void dragEnter( DropTargetEvent event ) {
        if (UiPlugin.isDebugging(Trace.DND)) {
            System.out.println("UDIGDropHandler.dragEnter: Setting event.detail to COPY"); //$NON-NLS-1$
        }
        event.detail = DND.DROP_COPY;
    }

	@Override
	public void dragOperationChanged( DropTargetEvent event ) {
        if (UiPlugin.isDebugging(Trace.DND)) {
            System.out.println("UDIGDropHandler.dragOperationChanged: Setting event.detail to COPY"); //$NON-NLS-1$
        }
        event.detail = DND.DROP_COPY;
    }
	
	@Override
	public void drop(DropTargetEvent event) {
		if (event.data != null) {
            this.type = event.data;
            this.transferType=event.dataTypes;
        }
	}
	
	public boolean validateDrop(Object target, int operation, TransferData transferType) {
		return true;
	}
	
	public boolean performDrop(Object data) {
        if (UiPlugin.isDebugging(Trace.DND)) {
            System.out.println("PerformDrop called on "+data+"("+data.getClass()+")"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        }
		//do a check for a multi object and seperate out
		Class type = data.getClass();
		Object[] objects = null;
		if (type.isArray()) {
			objects = (Object[])data;
		}
		else if (data instanceof Collection) {
			objects = ((Collection)data).toArray();
		}
		else if (data instanceof IStructuredSelection) {
			objects = ((IStructuredSelection)data).toArray();
		}
		else {
			objects = new Object[]{data};
		}

		CompositeDropActionJob queue = getJobQueue();
		
        //perform all the actions
		for (int i = 0; i < objects.length; i++) {
			Object object = objects[i];
			actions = UDIGDropProcessor.process(object, UDIGDropHandler.this);
			for (IDropAction action : actions) {
				queue.enqueue(new DropActionJob(object, this, action));
			}		
		}
		queue.schedule();
		
		return true; //assume actions were successful
	}
	
	private CompositeDropActionJob getJobQueue() {
		if (jobQueue == null) {
			synchronized (CompositeDropActionJob.class) {
				if (jobQueue == null) {
					jobQueue = new CompositeDropActionJob();
				}
			}
		}
		
		return jobQueue;
	}
	
	class CompositeDropActionJob extends Job {

		LinkedList<DropActionJob> queue;
		
		public CompositeDropActionJob() {
			super(""); //$NON-NLS-1$
			
			queue = new LinkedList<DropActionJob>();
		}

		public void enqueue(DropActionJob job) {
			synchronized (queue) {
				queue.addLast(job);
			}
		}
		
		@Override
		protected IStatus run(IProgressMonitor monitor) {
			DropActionJob next;
			while(true) {
				next = null;
				
				synchronized (queue) {
					if (!queue.isEmpty()) {
						next = queue.removeFirst();
					}
				}
				
				if (next == null) {
					//no more jobs to run
					return Status.OK_STATUS;
				}
				
				//run the next job
				next.run(monitor);
			}
		}
	}
	
	class DropActionJob extends Job {
		
		IDropAction action;
		UDIGDropHandler handler;
		Object data;
		
		public DropActionJob(Object data, UDIGDropHandler handler, IDropAction action) {
			super(action.getName());
			
			this.data = data;
			this.handler = handler;
			this.action = action;
		}
	
		@Override
		protected IStatus run(IProgressMonitor monitor) {
			try {
				action.perform(data, handler.getDestination(),monitor);
				return Status.OK_STATUS;
			}
			catch(Throwable t) {
				String msg = Policy.bind("UDIGDropHandler.error"); //$NON-NLS-1$
				String ns = action.getElement().getNamespace();
				
				Status s = new Status(IStatus.WARNING, ns, 0, msg, t);
				UiPlugin.getDefault().getLog().log(s);
				
				return Status.CANCEL_STATUS;
			}
		}
	}
}
