/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.internal.ui.operations;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import net.refractions.udig.core.internal.ExtensionPointList;
import net.refractions.udig.internal.ui.Policy;
import net.refractions.udig.internal.ui.UiPlugin;
import net.refractions.udig.ui.operations.OpAction;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.jface.action.GroupMarker;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.SubContributionItem;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.ui.IWorkbenchActionConstants;
import org.eclipse.ui.IWorkbenchWindow;

/**
 * Creates an Operation Sub menu containing operations for all 
 * operations defined for an object.
 * 
 * @author jeichar
 * @since 0.6.0
 */
public class OperationMenuFactory {
    ExtensionPointList extensionPoints;
    private MenuManager contextManager;
    Class targetClass;
    private IStructuredSelection selection;
    private Class actionClass;
	private ArrayList<OpAction> actions=new ArrayList<OpAction>();
	private Map<String, OperationCategory> categories=new HashMap<String, OperationCategory>();
	private MenuManager menuManager;
	private IWorkbenchWindow window;
	private final static Class<OpAction> ACTION_TYPE=OpAction.class;
    /**
	 * Create instance
	 */
	public OperationMenuFactory() {
		extensionPoints=new ExtensionPointList("net.refractions.udig.ui.operation"); //$NON-NLS-1$
	}
	

	/**
	 * Gets a context menu containing all the enabled operations
	 * 
	 * @param selection the current selection.
	 * @return a context menu containing all the enabled operations
	 */
    public MenuManager getContextMenu(){
        contextManager=new MenuManager(Policy.bind("OperationMenuFactory.1"), "analysis"); //$NON-NLS-1$ //$NON-NLS-2$
        for( OperationCategory category:categories.values() ){
        	MenuManager menu=category.createContextMenu();
        	if( menu.getItems().length>0 ){
        		contextManager.add(menu);
        	}
        	
        }
        for( OpAction action: actions){
        	if( action.isEnabled() ){
        		contextManager.add(action);
        	}
        }
        return contextManager;
    }

    /**
     * Creates a menu manager with actions that will enable or disable themselved
     * depending on the current workbench selection.
     * 
     * @param window the window that the created actions will listen to.
     * @return a menu manager with all the Operation Actions.
     */
    public MenuManager getMenu(IWorkbenchWindow window){
    	if (menuManager == null) {
    		this.window=window;
			menuManager = new MenuManager(Policy.bind("OperationMenuFactory.2")); //$NON-NLS-1$
	        createActionList(extensionPoints, window, ACTION_TYPE);
	        for( OperationCategory category:categories.values() ){
	        	if( category.getItems().length>0 ){
	        		menuManager.add(category);
                }
	        }
	        for( OpAction action: actions){
	        		menuManager.add(action);
	        }			
			menuManager.add(new GroupMarker(IWorkbenchActionConstants.MB_ADDITIONS));
		}
        return menuManager;
    }
    
    public void add(ExtensionPointList list, Class<OpAction> opActionClass){
    	menuManager.update();
    }

	private void createActionList(ExtensionPointList list, IWorkbenchWindow window, Class<OpAction> actionType) {
		for( IConfigurationElement element: list){
		    try{
		    	if( element.getName().equals("category") ) //$NON-NLS-1$
		    		categories.put(element.getAttribute("id"), new OperationCategory(element)); //$NON-NLS-1$
		    }catch (Exception e) {
                UiPlugin.log( null, e);
		    }
		}
		for( IConfigurationElement element: list){
			if( element.getName().equals("category")) //$NON-NLS-1$
				continue;
			OpAction action=new OpAction(element);
			if( window!=null )
				window.getSelectionService().addSelectionListener(action);
			OperationCategory category=categories.get(element.getAttribute("categoryId")); //$NON-NLS-1$
			if( category!=null)
				category.add(action);
			else
				actions.add(action);
		}
	}


    public void contributeActions( IMenuManager menuBar, IWorkbenchWindow window ) {
        for( OperationCategory category: categories.values() ){
            for( OpAction action : category.getActions() ) {
                addActionToMenu(menuBar, action);
            }
        }
        for(OpAction action: actions){
            addActionToMenu(menuBar, action);         
        }
    }


    /**
     * TODO summary sentence for addActionToMenu ...
     * 
     * @param menuBar
     * @param action
     */
    private void addActionToMenu( IMenuManager menuBar, OpAction action ) {
        if( action.getMenuPath()!=null )
            try{
                String[] paths=action.getMenuPath().split("/"); //$NON-NLS-1$
                IMenuManager manager = menuBar;
                String markerID=null;
                for( String path : paths ) {
                    markerID=null;
                    IContributionItem item=manager.findUsingPath(path);
                    if( item==null ){
                        UiPlugin.log(action.getMenuPath()+" is not a valid menuPath",null); //$NON-NLS-1$
                        break;
                    }else{
                        if( item instanceof IMenuManager ){
                            manager=(IMenuManager) item;
                        }else if( item.isGroupMarker() ){
                            markerID=item.getId();
                        }else if(item instanceof SubContributionItem){
                            item=((SubContributionItem)item).getInnerItem();
                            if( item instanceof IMenuManager ){
                                manager=(IMenuManager) item;
                            }else if( item.isGroupMarker()){
                                markerID=item.getId();
                            }
                        }
                    }
                }
                if(manager!=null){
                    if( markerID!=null )
                        manager.appendToGroup(markerID, action);
                    else{
                        manager.add( action );
                    }
                }else{
                    UiPlugin.log(action.getMenuPath()+" is not a valid menuPath",null); //$NON-NLS-1$                                
                }
            }catch (Exception e) {
                UiPlugin.log("Error adding operation to menu", e); //$NON-NLS-1$
            }
    }


}
