/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.location.internal;

import java.text.MessageFormat;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.resource.ImageDescriptor;

/**
 * Policy implements Internationalization convenience methods
 * for the plugin and makes progress monitor policy decisions.
 * 
 */
public class Policy {
	/** The resource bundle to get strings from */
	protected static ResourceBundle bundle = null;
	
	/**
	 * Creates a NLS catalog for the given locale.
	 * 
	 * @param bundleName  the name of the bundle
	 */
	public static void localize(String bundleName) {
		bundle = ResourceBundle.getBundle(bundleName);
	}
    
	/**
	 * Clean up after bundle.  
	 */
	public static void cleanup(){
        bundle = null;
	}
	
	/**
	 * Lookup the message with the given ID in this catalog and bind its
	 * substitution locations with the given string.
	 * 
	 * @param id  the id to look up
	 * @param binding  the string to bind to the result
	 * @return the bound string
	 */
	public static String bind(String id, Object binding) {
		return bind(id, new Object[] { binding });
	}
	
	/**
	 * Lookup the message with the given ID in this catalog and bind its
	 * substitution locations with the given strings.
	 * 
	 * @param id  the id to look up
	 * @param binding1  the first string to bind to the result
	 * @param binding2  the second string to bind to the result
	 * @return the bound string
	 */
	public static String bind(String id, Object binding1, Object binding2) {
		return bind(id, new Object[] { binding1, binding2 });
	}
	
	/**
	 * Lookup the message with the given ID in this catalog and bind its
	 * substitution locations with the given strings.
	 * 
	 * @param id  the id to look up
	 * @param binding1  the first string to bind to the result
	 * @param binding2  the second string to bind to the result
	 * @param binding3  the third string to bind to the result
	 * @return the bound string
	 */
	public static String bind(String id, Object binding1, Object binding2,Object binding3) {
		return bind(id, new Object[] { binding1, binding2, binding3 });
	}
	
	/**
	 * Gets a string from the resource bundle. We don't want to crash because of a missing String.
	 * Returns the key if not found.
	 * 
	 * @param key  the id to look up
	 * @param b nudle used for key binding
	 * @return the string with the given key
	 */
	public static String bind(String key, ResourceBundle b) {
		try {
			return b.getString(key);
		} catch (MissingResourceException e) {
			return key;
		} catch (NullPointerException e) {
			return "!" + key + "!"; //$NON-NLS-1$ //$NON-NLS-2$
		}
	}
	
	/**
	 * Gets a string from the resource bundle. We don't want to crash because of a missing String.
	 * Returns the key if not found.
	 * 
	 * @param key  the id to look up
	 * @return the string with the given key
	 */
	public static String bind(String key) {
		try {
			return bundle.getString(key);
		} catch (MissingResourceException e) {
			return key;
		} catch (NullPointerException e) {
			return "!" + key + "!"; //$NON-NLS-1$ //$NON-NLS-2$
		}
	}
	
	/**
	 * Gets a string from the resource bundle and binds it with the given arguments. If the key is 
	 * not found, return the key.
	 * 
	 * @param key  the id to look up
	 * @param args  the strings to bind to the result
	 * @return the bound string
	 */
	public static String bind(String key, Object[] args) {
		try {
			return MessageFormat.format(bind(key), args);
		} catch (MissingResourceException e) {
			return key;
		} catch (NullPointerException e) {
			return "!" + key + "!"; //$NON-NLS-1$ //$NON-NLS-2$
		}
	}
	
	/**
	 * Initialize the given Action from a ResourceBundle.
	 * <p>
	 * Makes use of the following keys:
	 * <ul>
	 * <li>prefix.label
	 * <li>prefix.tooltip
	 * <li>prefix.image
	 * <li>prefix.description
	 * </p>
	 * <p>Note: The use of a single image value is mapped to images
	 * for both the enabled and distabled state of the IAction.
	 * the Local toolbar (elcl16/ and dlcl16/) is assumed
	 * if a path has not been provided.
	 * <pre><code>
	 * add_co.gif              (prefix.image)
	 *    enabled: elcl16/add_co.gif
	 *   disabled: dlcl/remove_co.gif
	 * tool16/discovery_wiz.16 (prefix.image)
	 *    enabled: etool16/discovery_wiz.16
	 *   disabled: etool16/discovery_wiz.16
	 * </code></pre>
	 * </p>
	 * @param a action
	 * @param id used for binding (id.label, id.tooltip, ...)
	 */	
	public static void initAction(IAction a, String id ) {
		String labelKey = ".label"; //$NON-NLS-1$
		String tooltipKey = ".tooltip"; //$NON-NLS-1$
		String imageKey = ".image"; //$NON-NLS-1$
		String descriptionKey = ".description"; //$NON-NLS-1$
		if (id != null && id.length() > 0) {
			labelKey = id + labelKey;
			tooltipKey = id + tooltipKey;
			imageKey = id + imageKey;
			descriptionKey = id + descriptionKey;
		}
		String s = bind(labelKey);
		if (s != null)
			a.setText(s);
		s = bind(tooltipKey );
		if (s != null)
			a.setToolTipText(s);
		s = bind(descriptionKey );
		if (s != null)
			a.setDescription(s);
		String relPath = Policy.bind(imageKey );
		if (relPath != null && !relPath.equals(imageKey) && relPath.trim().length() > 0) {
			String dPath;
			String ePath;			
			if (relPath.indexOf("/") >= 0) { //$NON-NLS-1$
				String path = relPath.substring(1);
				dPath = 'd' + path;
				ePath = 'e' + path;
			} else {
				dPath = "dlcl16/" + relPath; //$NON-NLS-1$
				ePath = "elcl16/" + relPath; //$NON-NLS-1$
			}			
			ImageDescriptor image;
			
			image = Images.getDescriptor( ePath );
			if (id != null){
			    System.out.println( id+": '"+ ePath+"' found "+id );			     //$NON-NLS-1$ //$NON-NLS-2$
				a.setImageDescriptor(image);
			}
            image = Images.getDescriptor(dPath);			
			if (id != null){
                System.out.println( id+": '"+ dPath+"' found "+id );                 //$NON-NLS-1$ //$NON-NLS-2$
				a.setDisabledImageDescriptor(image);
			}
		}
	}
	
	/**
	 * Checks if the progress monitor is canceled.
	 * 
	 * @param monitor  the onitor to check for cancellation
	 * @throws OperationCanceledException if the monitor is canceled
	 */
	public static void checkCanceled(IProgressMonitor monitor) {
		if (monitor.isCanceled()) {
			throw new OperationCanceledException();
		}
	}
	/**
	 * Returns a monitor for the given monitor
	 * 
	 * @param monitor  the monitor to return a monitor for
	 * @return a monitor for the given monitor
	 */
	public static IProgressMonitor monitorFor(IProgressMonitor monitor) {
		if (monitor == null) {
			return new NullProgressMonitor();
		}
		return monitor;
	}	
	
	/**
	 * Set up a monitor - for a sub taks. 
	 * 
	 * @param monitor
	 * @param ticks
	 * @return ProgressMonitor of the provided number of ticks
	 */
	public static IProgressMonitor subMonitorFor(IProgressMonitor monitor, int ticks) {
		if (monitor == null)
			return new NullProgressMonitor();
		if (monitor instanceof NullProgressMonitor)
			return monitor;
		return new SubProgressMonitor(monitor, ticks);
	}
	
	/**
	 * Search backwards until split separators are found
	 * 
	 * @param path
	 * @param split
	 * @return like path shorter
	 */
	public static String toTruncatedPath(IPath path, int split) {
		int count = 0;
		String stringPath = path.toString();
		int index = stringPath.length();
		while (count++ < split && index != -1) {
			index = stringPath.lastIndexOf(IPath.SEPARATOR, index - 1);
		}
		if (index == -1) {
			return stringPath;
		}
		return "..." + stringPath.substring(index); //$NON-NLS-1$
	}

	public static ResourceBundle getBundle() {
		return bundle;
	}
}