/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.printing.ui;

import java.awt.Color;
import java.awt.Rectangle;
import java.text.MessageFormat;
import java.util.Locale;

import net.refractions.udig.mapgraphic.MapGraphic;
import net.refractions.udig.mapgraphic.MapGraphicContext;
import net.refractions.udig.project.ILayer;
import net.refractions.udig.ui.graphics.ViewportGraphics;

import org.geotools.geometry.GeneralDirectPosition;
import org.geotools.referencing.CRS;
import org.geotools.referencing.GeodeticCalculator;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.MathTransform;
import org.opengis.spatialschema.geometry.DirectPosition;

import com.vividsolutions.jts.geom.Coordinate;

/**
 * Provides a decorator object that represents a typical Scalebar
 * 
 * @author Richard Gould
 * @since 0.6.0
 */
public class ScalebarMapGraphic implements MapGraphic {

    /*
     * @see net.refractions.udig.project.render.decorator.Decorator#draw(java.awt.Graphics2D, net.refractions.udig.project.render.RenderContext)
     */
    public void draw( MapGraphicContext context) {
    	ViewportGraphics graphics = context.getGraphics();
        
        Rectangle location = (Rectangle) context.getLayer().getStyleBlackboard().get(LocationStyleContent.ID);
        if (location == null) {
            location = LocationStyleContent.createDefaultStyle();
            context.getLayer().getStyleBlackboard().put(LocationStyleContent.ID, location);
        }

        CoordinateReferenceSystem epsg4326;
        CoordinateReferenceSystem epsg4269;        
        try {
            epsg4326 = CRS.decode("EPSG:4326"); //$NON-NLS-1$
        } catch (FactoryException notFound) {
            epsg4326 = null;
        }
        try {
            epsg4269 = CRS.decode("EPSG:4269"); //$NON-NLS-1$
        }
        catch ( FactoryException notFound ){
            epsg4269 = null; // missing!
        }
        
        /*
         * 1. find out the width of the map [display] in pixels
         */
        
        int displayWidth = context.getRenderManagerInternal().getMapDisplay().getWidth();
        int displayHeight = context.getRenderManagerInternal().getMapDisplay().getHeight();
        
        /*
         * 2. take a coordinate from the left side and the right side of 
         * the map, in the map projection (whatever that may be).
         */        
        Coordinate minX = context.getViewportModelInternal().pixelToWorld(0, displayHeight/2);
        Coordinate maxX = context.getViewportModelInternal().pixelToWorld(displayWidth, displayHeight/2);
        
        /*
         * 3. back-project the coordinates from 2. to latlong (or, if 
         * they are already in latlong, leave them be)
         */        
        CoordinateReferenceSystem crs = context.getViewportModelInternal().getCRS();
        if( epsg4269 != null && epsg4269.equals( crs )){
            // already lat long - leave it alone
        } else if( epsg4326 != null && epsg4326.equals( crs )){
            // already lat long - leave it alone
        }
        else {            
            try {
                MathTransform transform;
                if ( epsg4269 != null ){
                    transform = CRS.transform(crs, epsg4269, true);
                }
                else if( epsg4326 != null ){
                    transform = CRS.transform(crs, epsg4326);
                }          
                else {
                    transform = CRS.transform(crs, DefaultGeographicCRS.WGS84);                    
                }                
                DirectPosition newMinX = transform.transform(new GeneralDirectPosition(minX.x, minX.y), null);
                DirectPosition newMaxX = transform.transform(new GeneralDirectPosition(maxX.x, maxX.y), null);
                
                minX = new Coordinate(newMinX.getOrdinate(0), newMinX.getOrdinate(1));
                maxX = new Coordinate(newMaxX.getOrdinate(0), newMaxX.getOrdinate(1));
            }
            catch (Exception e) {
                ILayer layer = context.getLayer();
                if( layer != null ){
                   // layer.setStatus( Layer.WARNING );                           
                }
                minX = null; // render as a
                maxX = null; // gray box for now
            }    
                        
        }
        
        /*
         * 4. calculate the distance between the two point, using the 
         * distance on a spheroid functions from the CRS packages
         */                
        double distance;
        
        if (minX == null || maxX == null){
            distance = 0.0;
        }
        else {
            try {
                GeodeticCalculator calc = new GeodeticCalculator();
                calc.setAnchorPoint(minX.x, minX.y);
                calc.setDestinationPoint(maxX.x, maxX.y);
                distance = calc.getOrthodromicDistance();
            }
            catch ( IllegalArgumentException outOfRange ){
                distance = 0.0;
            }
        }
        if (distance == 0.0 ){
            graphics.setColor(Color.GRAY);            
            graphics.fillRect( location.x, location.y, location.width, location.height );
            graphics.setColor(Color.BLACK);            
            //graphics.drawRect( location.x, location.y, location.width, location.height );
            graphics.draw( new Rectangle(location.x, location.y, location.width, location.height ));
            
            int yText = (location.y < (displayHeight/2)) ? location.y+location.height+15 : location.y-15;
            graphics.drawString(Policy.bind("ScalebarMapGraphic.zoomInRequiredMessage"), location.x, yText, -1, -1); //$NON-NLS-1$
            
            return;
        }
        /*
         * 5. now you have the width of the screen in pixels, and the 
         * width of the screen in meters, so you know the number of 
         * ground units per pixel
         */        
        double unitsPerPixel = distance/displayWidth; // will be 0 for unknown
        
        //System.out.println("Units per pixel: " + unitsPerPixel);
        
        /*
         * 6. judicious rounding routines will be required to get a 
         * "pretty" scalebar that has nice unts per color bar
         */        
        
        double endUnit = unitsPerPixel*location.width; // will be 0 for unknown
        
        //System.out.println("EndUNit: " + endUnit);
        //System.out.println(crs.getCoordinateSystem().getAxis(0).getUnit());
        //System.out.println(crs.getCoordinateSystem().getAxis(1).getUnit());
        
        /*
         * Draw the scale bar
         */
        int intervals = 5;
        
        int x = location.x;
        int y = location.y;
        
        int width = location.width/intervals;
        int height = location.height/2;
        
        double unitCount = 0;

        graphics.setColor(Color.BLACK);
        int yText = (location.y < (displayHeight/2)) ? location.y+location.height+15 : location.y-15;
        
        MessageFormat formatter = new MessageFormat("", Locale.getDefault()); //$NON-NLS-1$
        formatter.applyPattern(Policy.bind("ScalebarMapGraphic.numberOfKilometres")); //$NON-NLS-1$
        {
            Object[] arguments = { Math.round(unitCount)/1000 };
            String msg = formatter.format(arguments);

            //Rectangle2D text = graphics.getFontMetrics().getStringBounds( msg, graphics );
            //graphics.drawString( msg, (int) (x-(text.getWidth()/2)), yText );
            //graphics.drawString(Math.round(unitCount)+"", x+width, yText);
            graphics.drawString(msg, x, yText, -1, -1);
        }
        for (int i = 0; i < intervals; i++) {
        	
        	unitCount += endUnit/intervals;
        	
            if (i%2 == 0) {
                graphics.setColor(Color.BLACK);
            } else {
                graphics.setColor(Color.WHITE);
            }
            graphics.fillRect(x, y, width, height);
            
            graphics.setColor(Color.BLACK);
            {
                Object[] arguments = { Math.round(unitCount)/1000 };
                String msg = formatter.format(arguments);
                
                //Rectangle2D text = graphics.getFontMetrics().getStringBounds( msg, graphics );
                graphics.drawString(msg, x+width, yText, 0, -1);
            }            
            
            
            if (i%2 == 0) {
                graphics.setColor(Color.WHITE);
            } else {
                graphics.setColor(Color.BLACK);
            }
                        
            graphics.fillRect(x, y+height, width, height);
            
            x = x+width;            
        }        
    }

}