/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.project.command;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

import net.refractions.udig.project.internal.Policy;
import net.refractions.udig.project.internal.impl.MapImpl.MapCommandListener;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.util.ListenerList;

/**
 * TODO Purpose of net.refractions.udig.project.command <p> </p>
 * @author   Jesse
 * @since   1.0.0
 */
public class CommandManager implements CommandStack, NavCommandStack {

    ListenerList handlers = new ListenerList();
    ListenerList completionHandlers = new ListenerList();

    Executor commandExecutor = new Executor(Policy.bind("CommandManager.0"), this); //$NON-NLS-1$

    /**
     * Creates a new instance of CommandManager
     * 
     * @param handler an error handler to use to handle thrown exceptions.
     */
    public CommandManager( ErrorHandler handler, CommandListener commandCompletionListener ) {
        handlers.add(handler);
        completionHandlers.add(commandCompletionListener);
    }

    /**
     * Executes a command. Calls the Errorhandler if an exception is thrown.
     * 
     * @param command The command to execute
     * @param async flag indicating wether command should be executed sync vs async.
     */
    public void execute( final Command command, boolean async ) {
		int type = Request.RUN;
        Request request = new Request(type, command,async);
		commandExecutor.addCommand(request);
        if (request.isSync()) {
            //synchronous execution, current thread needs to wait
            // it is unlikley that the command will be complete by the time
            // we get here but better to be safe then sorry
            try {
                while (!request.completed ) {
                    synchronized (request) {
                        request.wait(1000);
                    }
                }
            }
            catch (InterruptedException e) {
                //TODO: log this
                e.printStackTrace();
            }
        }

	}

    /**
     * Executes the last undone command, if there are any commands to undo.
     */
    public void redo() {
        commandExecutor.addCommand(new Request(Request.REDO, null,true));
    }

    /**
     * Undoes the last command if possible.
     */
    public void undo() {
        commandExecutor.addCommand(new Request(Request.UNDO, null, true));
    }

    // /**
    // * Executes a command. Calls the Errorhandler if an exception is thrown.
    // *
    // * @param command The command to execute
    // */
    // public void execute(final Command command) {
    // try {
    // undone.clear();
    // command.run();
    // history.push(command);
    // } catch (Exception e) {
    // handleError(command, e);
    // }
    // }
    //
    // /**
    // * Executes the last undone command, if there are any commands to undo.
    // */
    // public void redo() {
    // if (undone.empty())
    // return;
    // Command command = (Command) undone.pop();
    // try {
    // command.run();
    // history.push(command);
    // } catch (Exception e) {
    // handleError(command, e);
    // }
    // }
    //
    // /**
    // * Undoes the last command if possible.
    // */
    // public void undo() {
    // if( history.isEmpty() )
    // return;
    // Command c = (Command) history.pop();
    // if (c instanceof UndoableCommand) {
    // UndoableCommand command = (UndoableCommand) c;
    // try {
    // command.rollback();
    // undone.push(command);
    // } catch (Exception e) {
    // handleRollbackError(command, e);
    // }
    // } else {
    // throw new RuntimeException(
    // "Undoing Commands (Not Undoable Command) is not handled " //$NON-NLS-1$
    // + "yet because it involves rolling back the current transaction and redoing all "
    // //$NON-NLS-1$
    // + "the commands in the stack"); //$NON-NLS-1$
    // }
    // }

    /**
     * Adds an Errorhandler to the list of error handlers
     * 
     * @param handler the error handler to add.
     * @see ErrorHandler
     */
    public void addErrorHandler( ErrorHandler handler ) {
        handlers.add(handler);
    }

    /**
     * Removes an Errorhandler from the list of error handlers
     * 
     * @param handler the error handler to remove.
     * @see ErrorHandler
     */
    public void removeErrorHandler( ErrorHandler handler ) {
        handlers.remove(handler);
    }

    /**
     * @see net.refractions.udig.project.command.CommandStack#canUndo()
     */
    public boolean canUndo() {
        Command c;
        if (!commandExecutor.history.isEmpty()) {
            c = (Command) commandExecutor.history.peek();
            if (c instanceof UndoableCommand)
                return true;
        }
        return false;
    }

    /**
     * @see net.refractions.udig.project.command.CommandStack#canRedo()
     */
    public boolean canRedo() {
        Command c;
        if (!commandExecutor.undone.isEmpty()) {
            c = (Command) commandExecutor.undone.peek();
            if (c instanceof UndoableCommand)
                return true;
        }
        return false;
    }

    /**
     * @see net.refractions.udig.project.command.NavCommandStack#hasBackHistory()
     */
    public boolean hasBackHistory() {
        return canUndo();
    }

    /**
     * @see net.refractions.udig.project.command.NavCommandStack#hasForwardHistory()
     */
    public boolean hasForwardHistory() {
        return canRedo();
    }

    /**
     * TODO Purpose of net.refractions.udig.project.command <p> </p>
     * @author   Jesse
     * @since   1.0.0
     */
    public class Executor extends Job {
        Stack<Command> history = new Stack<Command>();

        Stack<Command> undone = new Stack<Command>();

        List<Request> commands=new LinkedList<Request>();
        
        CommandManager parent;
        /**
         * Construct <code>Executor</code>.
         * 
         * @param name the name of the job
         * @param parent the CommandManager instance that is calling the executor
         * @param type the type of the executor. (RUN, UNDO, REDO)
         */
        public Executor( String name, CommandManager parent ) {
            super(name);
            this.parent = parent;
            setSystem(true);
            setPriority(LONG);
        }
        IProgressMonitor current;

        @Override
        protected IStatus run( IProgressMonitor monitor ) {
        	current=monitor;
            Request request;
            while( !commands.isEmpty() ) {
                synchronized (this) {
                    request = commands.get(0);
                    commands.remove(0);
                }
                run(monitor, request);
				
				if (request.isSync()) {
					//notify those wating for command to finish
					synchronized (request) {
						request.notifyAll();
					}
				}
            }
			
			return Status.OK_STATUS;
        }
		private void run(IProgressMonitor monitor, Request request) {
			switch( request.type ) {
			case Request.RUN:
			    execute(request.command, monitor);
			    break;
			case Request.UNDO:
			    undo(monitor);
			    break;
			case Request.REDO:
			    redo(monitor);
			    break;
			case Request.RERUN:
			    rerunCommands(monitor);
			    break;
			}

			request.completed = true;
		}

        /**
         * Adds a command to the stack of commands that needs to be executed.
         * 
         * @param request
         */
        public synchronized void addCommand( Request request ) {
        	if( getThread()==Thread.currentThread()){
        		run(current, request);
        	}else{
				commands.add(request);
				schedule();
        	}
        }

        /**
         * Executes a command. Calls the Errorhandler if an exception is thrown.
         * 
         * @param command The command to execute
         */
        private void execute( final Command command, IProgressMonitor monitor) {
            try {
                undone.clear();
                command.run(monitor);
                history.push(command);
                notifyOwner(command);
            } catch (Exception e) {
                handleError(command, e);
            }
        }

		/**
		 * Notifies the owner that the command has been executed.
		 */
        private void notifyOwner( Command command ) {
            Object[] harray = completionHandlers.getListeners();
            for( int i = 0; i < harray.length; i++ ) {
                if (command instanceof NavCommand) {
                    ((CommandListener) harray[i]).commandExecuted(MapCommandListener.NAV_COMMAND);
                } else {
                    ((CommandListener) harray[i]).commandExecuted(MapCommandListener.COMMAND);
                }
            }
        }

        /**
         * Executes the last undone command, if there are any commands to undo.
         */
        private void redo( IProgressMonitor monitor ) {
            Command command=undone.pop();
            try {
                command.run(monitor);
                history.push(command);
                notifyOwner(command);
            } catch (Exception e) {
                handleError(command, e);
            }
        }

        /**
         * Undoes the last command if possible.
         */
        private void undo( IProgressMonitor monitor ) {
            Command c;
            //First check if there's a command on the commands stack that hasn't 
            //been executed and should just be removed
            if (!commands.isEmpty()) {
            	Request r;
            	synchronized (this) {
                    r = commands.get(0);
                    commands.remove(0);
				}
                c = r.command;
                if (c instanceof UndoableCommand) {
                    //We've already popped it off...so we're done
                } else {
                    throw new RuntimeException(
                            "Undoing Commands (No Undoable Command) is not handled " //$NON-NLS-1$
                                    + "yet because it involves rolling back the current transaction and redoing all " //$NON-NLS-1$
                                    + "the commands in the stack"); //$NON-NLS-1$
                }
                return;
            }
            //Nothing on the Command stack, we'll actually undo something
            c = history.pop();
            if (c instanceof UndoableCommand) {
                UndoableCommand command = (UndoableCommand) c;
                try {
                    command.rollback(monitor);
                    undone.push(command);
                } catch (Exception e) {
                    handleRollbackError(command, e);
                }
            } else {
                throw new RuntimeException(
                        "Undoing Commands (No Undoable Command) is not handled " //$NON-NLS-1$
                                + "yet because it involves rolling back the current transaction and redoing all " //$NON-NLS-1$
                                + "the commands in the stack"); //$NON-NLS-1$
            }
            notifyOwner(c);
        }

        private void handleError( Command command, Exception e ) {
            Object[] harray = handlers.getListeners();
            for( int i = 0; i < harray.length; i++ ) {
                ((ErrorHandler) harray[i]).handleError(command, e);
            }
        }

        private void handleRollbackError( UndoableCommand command, Exception e ) {
            Object[] harray = handlers.getListeners();
            for( int i = 0; i < harray.length; i++ ) {
                ((ErrorHandler) harray[i]).handleRollbackError(command, e);
            }
        }

        /**
         * Executes all the commands in the history again.
         */
        public void rerunCommands(IProgressMonitor monitor) {
            Stack<Command> h = history;
            history = new Stack<Command>();
            for( Iterator<Command> iter = h.iterator(); iter.hasNext(); ) {
                Command command = (Command) iter.next();
                execute(command, monitor);
            }
        }
    }
    /**
     * TODO Purpose of net.refractions.udig.project.command <p> </p>
     * @author   Jesse
     * @since   1.0.0
     */
    public class Request {
        /** <code>RUN</code> field */
        public static final int RUN = 0;
        /** <code>UNDO</code> field */
        public static final int UNDO = 1;
        /** <code>REDO</code> field */
        public static final int REDO = 2;
        /** <code>RERUN</code> field */
        public static final int RERUN = 4;
		
        /** the type of request */
        public int type;
		/** sync/async **/
		public boolean async;
		/** the command to be done/undone/redone */
        public Command command;
		/** flag to signal wether command is complete **/
		public boolean completed;
		
        /**
         * Construct <code>Request</code>.
         * 
         * @param type the type of request
         * @param command the command to be done/undone/redone
         */
        public Request( int type, Command command, boolean async ) {
            this.command = command;
            this.type = type;
			this.async = async;
			completed = false;
        }
        /**
         * Construct <code>Request</code>.
         * 
         * @param type the type of request
         */
        public Request( int type ) {
            this.type = type;
        }
		
		/**
		 * Determines if the request is synchronous.
		 * @return
		 */
		public boolean isSync() {
			return !async;
		}
    }

}