/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.project.command;

import net.refractions.udig.project.ILayer;
import net.refractions.udig.project.internal.Layer;
import net.refractions.udig.project.internal.commands.DeleteLayerCommand;
import net.refractions.udig.project.internal.commands.edit.CreateFeatureCommand;
import net.refractions.udig.project.internal.commands.edit.DeleteFeatureCommand;
import net.refractions.udig.project.internal.commands.edit.ModifyEditFeatureCommand;
import net.refractions.udig.project.internal.commands.edit.NullEditFeatureCommand;
import net.refractions.udig.project.internal.commands.edit.ResetEditFeatureCommand;
import net.refractions.udig.project.internal.commands.edit.RollbackCommand;
import net.refractions.udig.project.internal.commands.edit.SetAttributeCommand;
import net.refractions.udig.project.internal.commands.edit.SetGeomteryCommand;
import net.refractions.udig.project.internal.commands.edit.WriteEditFeatureCommand;
import net.refractions.udig.project.internal.commands.selection.CommitCommand;

import org.geotools.feature.Feature;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;

/**
 * Creates Edit commands which must be used to modify editable feature data.
 * 
 * API internal classes are in the returned API
 * 
 * @author jeichar
 * @since 0.3
 */
public class EditCommandFactory {
	/**
	 * Creates a new EditCommandFactory object
	 * 
	 * @return a new EditCommandFactory object
	 */
	public static EditCommandFactory getInstance() {
		return instance;
	}
	
	private static final EditCommandFactory instance = new EditCommandFactory();
	private EditCommandFactory(){
		// no op
	}

	/**
	 * Creates a {@linkplain SetAttributeCommand}object.
	 * 
	 * @param xpath xpath that identifies an attribute in the current editable Feature
	 * @param value the value that the attribute will be set to.
	 * @return a new {@linkplain SetAttributeCommand}object
	 * @see EditCommand
	 */
	public EditCommand createSetAttributeCommand(String xpath, Object value) {
		return new SetAttributeCommand(xpath, value);
	}

	/**
	 * Creates a {@linkplain SetGeomteryCommand}object.
	 * 
	 * @param xpath xpath that identifies an attribute in the current editable Feature
	 * @param geom the geom (in <b>layer </b> CRS) that the geometry will be set to.
	 * @return a new {@linkplain SetGeomteryCommand}object
	 * @see EditCommand
	 * @see Geometry
	 */
	public EditCommand createSetGeomteryCommand(String xpath, Geometry geom) {
		return new SetGeomteryCommand(xpath, geom);
	}

	/**
	 * Creates a {@linkplain SetGeomteryCommand}object that sets the default geometry.
	 * 
	 * @param geom the geom (in <b>layer </b> CRS) that the geometry will be set to.
	 * @return a new {@linkplain SetGeomteryCommand}object.
	 * @see EditCommand
	 * @see Geometry
	 */
	public EditCommand createSetGeomteryCommand(Geometry geom) {
		return new SetGeomteryCommand(SetGeomteryCommand.DEFAULT, geom);
	}

	/**
	 * Creates a {@linkplain NullEditFeatureCommand}object that sets the current editVictim victim.
	 * 
	 * @param feature the feature that will be the new editable Feature.
	 * @param layer A victim Store that contains the editable Feature.
	 * @return a new {@linkplain NullEditFeatureCommand}object.
	 * @see Feature
	 * @see Layer
	 * @see EditManagerControlCommand
	 */
	public MapCommand createSetEditFeatureCommand(
			Feature feature, ILayer layer) {
		return new NullEditFeatureCommand(feature, layer);
	}
    /**
     * Create a Commit command
     * @return a new {@linkplain CommitCommand} object that deletes the feature.
	 * @see CommitCommand
     */
    public MapCommand createCommitCommand() {
        return new CommitCommand();
    }
    /**
     * Create a Commit command
     * @return a new {@linkplain RollbackCommand} object that deletes the feature.
	 * @see RollbackCommand
     */
    public MapCommand createRollbackCommand() {
        return new RollbackCommand();
    }

	/**
	 * Creates a {@linkplain NullEditFeatureCommand}object that sets the current editable Feature.
	 * 
	 * @param feature the feature that will be the new editable Feature.
	 * @return a new {@linkplain NullEditFeatureCommand}object.
	 * @see EditManagerControlCommand
	 * @see Feature
	 */
	public MapCommand createSetEditFeatureCommand(Feature feature) {
		return new NullEditFeatureCommand(feature);
	}

	/**
	 * Creates a {@linkplain NullEditFeatureCommand}object that sets the current editable Feature to
	 * null.
	 * 
	 * @return a new {@linkplain NullEditFeatureCommand}object that sets the current editable
	 *         Feature to null..
	 * @see EditManagerControlCommand
	 */
	public MapCommand createNullEditFeatureCommand() {
		return new NullEditFeatureCommand(null, null);
	}

	/**
	 * @param coordinates the coordinates of the new feature in <b>Map </b> CRS.
	 * @return a new {@linkplain CreateFeatureCommand}object creates the feature.
	 * @see CreateFeatureCommand
	 */
	public MapCommand createFeature(Coordinate[] coordinates) {
		return new CreateFeatureCommand(coordinates);
	}

	/**
	 * Create a delete layer command
	 * 
	 * @param map the map containing the layer
	 * @param layer the layer to delete
	 * @return a new {@linkplain DeleteLayerCommand}object that deletes the layer.
	 * @see DeleteLayerCommand
	 */
	public MapCommand createDeleteLayer(Layer layer) {
		return new DeleteLayerCommand(layer);
	}

	/**
	 * Create a Delete Feature command
	 * 
	 * @param layer the layer containing the feature
	 * @param feature the feature to delete
	 * @return a new {@linkplain DeleteFeatureCommand}object that deletes the feature.
	 * @see DeleteFeatureCommand
	 */
	public MapCommand createDeleteFeature(Feature feature, ILayer layer) {
		return new DeleteFeatureCommand(feature, (Layer) layer);
	}

	/**
	 * Create a {@linkplain WriteEditFeatureCommand} command
	 * 
	 * @return a new {@linkplain WriteEditFeatureCommand} object that deletes the feature.
	 * @see WriteEditFeatureCommand
	 */
	public MapCommand createWriteEditFeatureCommand() {
		return new WriteEditFeatureCommand();
	}
	/**
	 * Create a {@linkplain ResetEditFeatureCommand} command
	 * 
	 * @return a new {@linkplain ResetEditFeatureCommand} object that deletes the feature.
	 * @see ResetEditFeatureCommand
	 */
	public MapCommand createResetEditFeatureCommand() {
		return new ResetEditFeatureCommand();
	}
    /**
     * Create a {@linkplain ModifyEditFeatureCommand} command
     * 
     * @return a new {@linkplain ModifyEditFeatureCommand} object that deletes the feature.
     * @see ModifyEditFeatureCommand
     */
    public MapCommand createModifyEditFeatureCommand( Feature editedFeature ) {
        return new ModifyEditFeatureCommand(editedFeature);
    }

}
