/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.project.ui.internal.actions;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import net.refractions.udig.project.command.EditCommandFactory;
import net.refractions.udig.project.internal.Layer;
import net.refractions.udig.project.internal.Project;
import net.refractions.udig.project.internal.ProjectElement;
import net.refractions.udig.project.internal.ProjectPlugin;
import net.refractions.udig.project.ui.UDIGGenericAction;
import net.refractions.udig.project.ui.commands.DrawCommandFactory;
import net.refractions.udig.project.ui.internal.Policy;
import net.refractions.udig.project.ui.internal.ProjectUIPlugin;
import net.refractions.udig.project.ui.internal.UDIGEditorInputDescriptor;
import net.refractions.udig.project.ui.internal.commands.draw.DrawFeatureCommand;
import net.refractions.udig.project.ui.render.displayAdapter.ViewportPane;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.MessageDialogWithToggle;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PlatformUI;
import org.geotools.feature.Feature;

/**
 * Deletes the selected elements from a project.
 * 
 * @author jeichar
 * @since 0.3
 */
public class Delete extends UDIGGenericAction {

	/**
	 * @see net.refractions.udig.project.ui.UDIGGenericAction#operate(net.refractions.udig.project.Layer)
	 */
	protected void operate(Layer layer) {
		layer.getContextModel().getMap().sendCommandASync(
				EditCommandFactory.getInstance().createDeleteLayer(layer));
	}

	/**
	 * @see net.refractions.udig.project.ui.UDIGGenericAction#operate(net.refractions.udig.project.IProjectElement)
	 */
	protected void operate(ProjectElement element) {
		doDelete(element, false, Window.OK);
	}

	protected final void doDelete(ProjectElement element, boolean deleteProjectFiles, int returncode) {
		for (UDIGEditorInputDescriptor desc : ProjectUIPlugin.getDefault()
				.getEditorInputs(element.getClass())) {
			try {
				IWorkbenchPage page = PlatformUI.getWorkbench()
						.getActiveWorkbenchWindow().getActivePage();
				IEditorPart editor = page.findEditor(desc.createInput(element));
				if (editor != null)
					page.closeEditor(editor, false);
			} catch (CoreException e) {
				// fall through
			}
		}
		element.getProjectInternal().getElementsInternal().remove(element);
		Resource resource=element.eResource();
		if( resource!=null ){
			resource.getContents().remove(element);
			resource.unload();
		}
	}

	/**
	 * @see net.refractions.udig.project.ui.UDIGGenericAction#operate(net.refractions.udig.project.Project)
	 */
	protected void operate(Project project) {

		MessageDialogWithToggle dialog = MessageDialogWithToggle
				.openOkCancelConfirm(Display.getDefault().getActiveShell(),
						"Delete Project", "Delete Project \""
								+ project.getName() + "\"?",
						"Delete from file system", false, null, null);

		boolean deleteProjectFiles=dialog.getToggleState();
		doDelete(project, deleteProjectFiles,dialog.getReturnCode());
	}

	protected final void doDelete(Project project, boolean deleteProjectFiles, int returncode) {
		if (returncode != Window.CANCEL) {
			Resource resource = project.eResource();
			if (!deleteProjectFiles) {
				try {
					resource.save(null);
					resource.getContents().remove(project);
				} catch (IOException e) {
					ProjectUIPlugin.log(null, e);
				}
			}
			ProjectPlugin.getPlugin().getProjectRegistry().getProjects()
					.remove(project);
			ProjectPlugin.getPlugin().getProjectRegistry().setCurrentProject(
					null);
			List<ProjectElement> toRemove = new ArrayList<ProjectElement>();
			toRemove.addAll(project.getElementsInternal());
			for (ProjectElement element : toRemove) {
				resource.setModified(false);
				operate(element);
			}
			resource.unload();
			if (deleteProjectFiles) {
				try {
					resource.getResourceSet().getResources().remove(resource);
					String path = resource.getURI().toFileString();
					path = path.substring(0, path.lastIndexOf('/'));
					final File file = new File(path);
					if (file.exists()) {
						String[] files = file.list(new FilenameFilter(){

							public boolean accept(File dir, String name) {
								if ( dir==file && name.endsWith(".udig"))
									return true;
								return false;
							}
							
						});
						for (String string : files) {
							File temp=new File(string);
							temp.delete();
						}
						file.delete();
					}

				} catch (Exception e) {
					ProjectUIPlugin.log("Error creating Project File", e);
				}
			}
		}
	}

	/**
	 * @see net.refractions.udig.project.ui.UDIGGenericAction#operate(org.geotools.feature.Feature)
	 */
	protected void operate(final Feature feature) {
		IAdaptable adaptableFeature = null;
		if (feature instanceof IAdaptable) {
			adaptableFeature = (IAdaptable) feature;
		}
		if (adaptableFeature == null) {
			adaptableFeature = (IAdaptable) Platform.getAdapterManager()
					.getAdapter(feature, IAdaptable.class);
		}
		if (adaptableFeature == null)
			return;

		final Layer layer = (Layer) adaptableFeature.getAdapter(Layer.class);

		DrawFeatureCommand command = DrawCommandFactory.getInstance()
				.createDrawFeatureCommand(feature, layer);

		ViewportPane pane = (ViewportPane) layer.getMapInternal()
				.getRenderManager().getMapDisplay();
		pane.addDrawCommand(command);

		PlatformUI.getWorkbench().getDisplay().syncExec(new Runnable() {
			public void run() {
				boolean result = MessageDialog.openConfirm(PlatformUI
						.getWorkbench().getDisplay().getActiveShell(), Policy
						.bind("DeleteFeature.confirmation.title"), //$NON-NLS-1$
						Policy.bind("DeleteFeature.confirmation.text")); //$NON-NLS-1$

				if (result) {
					layer.getContextModel().getMap().sendCommandASync(
							EditCommandFactory.getInstance()
									.createDeleteFeature(feature, layer));
				}
			}
		});
		command.setValid(false);
		pane.repaint();

	}

}
