/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.style.sld;

import java.awt.Color;
import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.net.URL;
import java.util.Random;

import javax.xml.transform.TransformerException;

import net.refractions.udig.catalog.IGeoResource;
import net.refractions.udig.catalog.IService;
import net.refractions.udig.project.StyleContent;
import net.refractions.udig.project.internal.Layer;
import net.refractions.udig.ui.graphics.SLDs;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.ui.IMemento;
import org.geotools.data.FeatureSource;
import org.geotools.data.shapefile.ShapefileDataStore;
import org.geotools.feature.FeatureType;
import org.geotools.feature.GeometryAttributeType;
import org.geotools.styling.FeatureTypeStyle;
import org.geotools.styling.Fill;
import org.geotools.styling.LineSymbolizer;
import org.geotools.styling.PointSymbolizer;
import org.geotools.styling.PolygonSymbolizer;
import org.geotools.styling.RasterSymbolizer;
import org.geotools.styling.SLDParser;
import org.geotools.styling.SLDTransformer;
import org.geotools.styling.Stroke;
import org.geotools.styling.Style;
import org.geotools.styling.StyleBuilder;
import org.geotools.styling.StyleFactory;
import org.geotools.styling.TextSymbolizer;


/**
 * Style content for Style Layer Descriptor (SLD).
 * This class is not intended for extention (jg: so I marked it final).
 * 
 * @author Justin Deoliveira, Refractions Research Inc.
 */
public final class SLDContent extends StyleContent {

    /** style id, used to identify sld style on a blackboard * */
    public static final String ID = "net.refractions.udig.style.sld"; //$NON-NLS-1$

    /** factory used to create style and builder * */
    private static StyleFactory styleFactory = StyleFactory.createStyleFactory();

    /** factory used to create style content * */
    private static StyleBuilder styleBuilder = new StyleBuilder(styleFactory);

    /** fandom number generator used to create random colors * */
    private static Random random = new Random();

    /**
     * SLDContent constructor.
     */
    public SLDContent() {
        super(ID);
    }

    /*
     * (non-Javadoc)
     * 
     * @see net.refractions.udig.project.StyleContent#getStyleClass()
     */
    public Class getStyleClass() {
        return Style.class;
    }

    /*
     * (non-Javadoc)
     * 
     * @see net.refractions.udig.project.StyleContent#save(org.eclipse.ui.IMemento,
     *      java.lang.Object)
     */
    public void save( IMemento momento, Object value ) {
        Style style = (Style) value;

        // serialize out the style objects
        SLDTransformer sldWriter = new SLDTransformer();
        String out = ""; //$NON-NLS-1$
        try {
            out = sldWriter.transform(style);
        } catch (TransformerException e) {
            SLDPlugin.log(null, e);
            e.printStackTrace();
        } catch (Exception e) {
            SLDPlugin.log(null, e);
        }
        momento.putTextData(out);
        momento.putString("type", "SLDStyle"); //$NON-NLS-1$ //$NON-NLS-2$
        momento.putString("version", "1.0"); //$NON-NLS-1$ //$NON-NLS-2$
    }

    /*
     * (non-Javadoc)
     * 
     * @see net.refractions.udig.project.StyleContent#load(org.eclipse.ui.IMemento)
     */
    public Object load( IMemento momento ) {
        // parse the sld object
        StringReader reader = new StringReader(momento.getTextData());
        SLDParser sldParser = new SLDParser(getStyleFactory(), reader);

        Style[] parsed = sldParser.readXML();
        if (parsed != null && parsed.length > 0)
            return parsed[0];

        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see net.refractions.udig.project.StyleContent#load(java.net.URL)
     */
    public Object load( URL url, IProgressMonitor m ) throws IOException {
        return parse(url);
    }

    public static void apply(Layer layer, Style style, IProgressMonitor m)
    	throws IOException {
    	
    	if (layer == null) return;
    	if (style == null) return;
    		
    	if (layer.isType(FeatureSource.class)) {
    		IGeoResource resource = layer.getGeoResource(FeatureSource.class);
    		FeatureSource featureSource 
    			= resource.resolve(FeatureSource.class, m);
    		
    		if (featureSource != null) {
    			//match up the feature type style name and the feature type name
    			FeatureType type = featureSource.getSchema();
    			FeatureTypeStyle fstyle = SLDs.featureTypeStyle(style,type);
    			if (fstyle == null) {
    				//force a name match
    				FeatureTypeStyle[] fstyles = style.getFeatureTypeStyles();
    				if (fstyles != null && fstyles.length > 0) {
    					fstyle = fstyles[0];
    				}
    			}
    			
    			if (fstyle != null) {
    				fstyle.setName(type.getTypeName());
    				layer.getStyleBlackboard().put(SLDContent.ID, style);
//    				
//    				//force a rerender, TODO: blackboard events
//    				layer.getMap().getRenderManager().refresh(
//    					layer, resource.getInfo(m).getBounds()	
//    				);
    			}
    		}
    	}
    }
    
    /**
     * This will need to know the "scheme."
     */
    public Object createDefaultStyle( IGeoResource resource, Color colour, 
            IProgressMonitor m ) throws IOException {

        if (resource.canResolve(FeatureSource.class)) {
            FeatureSource featureSource = null;
            try {
                featureSource = resource.resolve(FeatureSource.class, m);
            } 
            catch (IOException e) {
                e.printStackTrace();
                return null;
            }
            
            if (featureSource == null) {
                return null;
            }
            
            FeatureType schema = featureSource.getSchema();
            Style style = null;
            
            // do a check for shapefiles, there may be a parallel sld
            IService service = resource.service(m);
            if (service.canResolve(ShapefileDataStore.class)) {
            	URL url = service.getIdentifier();
            	String shp = url.getFile();
            	
            	//strip off the extension and check for sld
            	String sld = shp.substring(0, shp.length()-4) + ".sld";
            	File f = new File(sld);
            	if (!f.exists()) {
            		//try upper case
            		sld = shp.substring(0, shp.length()-4) + ".SLD";
            		f = new File(sld);
            	}
            	
            	if (f.exists()) {
            		//parse it up
            		SLDParser parser = new SLDParser(styleFactory);
            		parser.setInput(f);
            		Style[] styles = parser.readXML();
            		
            		//put the first one on
            		if (styles != null && styles.length > 0) {
            			style = SLDs.matchingStyle(styles,schema);
            			if (style == null) {
            				//just choose the first one
            				style = styles[0];
            			}
            		}
            	}
            }
            
            SLDContentManager sldContentManager 
            	= new SLDContentManager(styleBuilder, style);
            
            
            if (style == null) {
            	//fall back to create some default
            	style = styleBuilder.createStyle();
            	sldContentManager = new SLDContentManager(styleBuilder, style);
            	
            	//initialize the symbolizers based on geometry type
            	GeometryAttributeType geom = schema.getDefaultGeometry();
            	if ( SLD.isLine( geom )) {
                    sldContentManager.addSymbolizer(createLineSymbolizer(colour));
                }
                else if ( SLD.isPoint( geom )) {
                    sldContentManager.addSymbolizer(styleBuilder.createPointSymbolizer());
                }
                else if ( SLD.isPolygon( geom )) {
                    sldContentManager.addSymbolizer(createPolygonSymbolizer(colour));
                }
                else { 
                	// XXX: must be generic Geometry?
                	sldContentManager.addSymbolizer(styleBuilder.createPointSymbolizer());
                	sldContentManager.addSymbolizer(styleBuilder.createLineSymbolizer());
                	
                }
            }

            //set the feature type name
            sldContentManager.getDefaultFeatureTypeStyle()
            	.setFeatureTypeName(schema.getTypeName());

            return style;
        }
        return null;
    }
    
    public static Style parse(URL url) throws IOException {
    	StyleFactory factory = StyleFactory.createStyleFactory();
        SLDParser styleReader = new SLDParser(factory, url);
        Style style = styleReader.readXML()[0];

        return style;
    }
    
    public static Style createDefaultStyle() {
        Style style = styleBuilder.createStyle();
        SLDContentManager sldContentManager = new SLDContentManager(styleBuilder, style);

        // sldContentManager.addSymbolizer(styleBuilder.createPointSymbolizer());
        sldContentManager.addSymbolizer(createLineSymbolizer(createRandomColor()));
        sldContentManager.addSymbolizer(createPolygonSymbolizer(createRandomColor()));
        sldContentManager.addSymbolizer(createTextSymbolizer());
        // sldContentManager.addSymbolizer(styleBuilder.createRasterSymbolizer());

        return style;
    }

    public static StyleFactory getStyleFactory() {
        return styleFactory;
    }

    public static StyleBuilder getStyleBuilder() {
        return styleBuilder;
    }

    protected static PointSymbolizer createPointSymbolizer(Color colour) {
        PointSymbolizer symbolizer = styleBuilder.createPointSymbolizer();
        return symbolizer;
    }

    protected static LineSymbolizer createLineSymbolizer(Color colour) {
//    	 added Color handling 
		// Pati 2005-03-08
		if(colour == null) {
			colour = createRandomColor();
		}
        LineSymbolizer symbolizer = styleBuilder.createLineSymbolizer();
       
        
        Stroke stroke = styleBuilder.createStroke();
        stroke.setColor(styleBuilder.colorExpression(colour));
        stroke.setWidth(styleBuilder.literalExpression(1));
        symbolizer.setStroke(stroke);

        return symbolizer;
    }

    protected static PolygonSymbolizer createPolygonSymbolizer(Color colour) {
//   	 added Color handling 
		// Pati 2005-03-08
    	if(colour == null) {
			colour = createRandomColor();
		}
        PolygonSymbolizer symbolizer = styleBuilder.createPolygonSymbolizer();

        Stroke stroke = styleBuilder.createStroke();
        stroke.setColor(styleBuilder.colorExpression(colour));
        stroke.setWidth(styleBuilder.literalExpression(1));
        symbolizer.setStroke(stroke);

        Fill fill = styleBuilder.createFill();
        fill.setColor(styleBuilder.colorExpression(colour));
        fill.setOpacity(styleBuilder.literalExpression(.5));
        symbolizer.setFill(fill);

        return symbolizer;
    }

    protected static TextSymbolizer createTextSymbolizer() {
        TextSymbolizer symbolizer = styleBuilder.createTextSymbolizer();
        return symbolizer;
    }

    protected static RasterSymbolizer createRasterSymbolizer() {
        RasterSymbolizer symbolizer = styleBuilder.createRasterSymbolizer();
        return symbolizer;
    }

    protected static Color createRandomColor() {
        return new Color(random.nextInt(200), random.nextInt(200), random.nextInt(200));
    }
}