package net.refractions.udig.ui;

import java.io.IOException;

import net.refractions.udig.internal.ui.Policy;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.part.PageBook;
import org.geotools.data.FeatureReader;
import org.geotools.data.FeatureResults;
import org.geotools.feature.AttributeType;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureCollections;
import org.geotools.feature.FeatureType;

/**
 * 
 * A TreeViewer control for viewing a table of Feature attributes. 
 * <p>
 * This object can be created in two ways. The first is using a FeatureReader. 
 * This method reads in Features a page at a time and populates the table 
 * entries. The page size can be set by calling FeatureTableControl#setPageSize(int).
 * </p>
 * <p>
 * The second way is using a FeatureCollection. In this case the control hangs
 * on to a reference to the FeatureCollection and populates the table entries
 * directory from it. This method results in a single page containing all 
 * features.
 * </p>
 * TODO: Create next and first action
 * @author jdeolive
 * @since 0.3
 */
public class FeatureTableControl  {
    
    /** results per page **/
    private int pageSize = 10; // XXX: actual put this as a user pref
    
    /** feature types and reader to get the actual feature data **/
    private FeatureReader fReader;
    private FeatureType fType;
    
    /** feature collection for holding features in memory **/
    private FeatureCollection features;
    
    /** table viewer control **/
    private TableViewer tableViewer;

    private FeatureResults results;

    private PageBook book;

    /** Action to move the table to the first Page */
    public IAction first;
    
    /** Action to move to the previous Page */
    public IAction prev;
    
    /** Action to move to the next Page */
    public IAction next;
    
    /** Action to move to the lastPage */
    public IAction last;

    private Text message;
    
   
    /**
     * Construct <code>FeatureTableControl</code>.
     * <p>
     * Must call setFeatures before use.
     * </p>
     */
    public FeatureTableControl() {
        clear();        
    }
    /**
     * 
     * Construct <code>FeatureTableControl</code>.
     * @param parent 
     *
     * @param features The FeatureCollection containing the features to be
     * displayed.
     */
    public FeatureTableControl(Composite parent, FeatureCollection features) {
        createTableControl( parent );
        setFeatures(features);
    }
    
    /**
     * 
     * Construct a <code>FeatureTableControl</code>.
     *
     * @param fReader The FeatureReader that returns the actual features.
     */
    public FeatureTableControl(Composite parent, FeatureReader fReader) {
        this( parent, fReader,10);
    }
    
    /**
     * 
     * Construct a <code>FeatureTableControl</code>.
     *
     * @param fReader The FeatureReader that returns the actual features.
     * @param resPerPage Results per page to be shown in the table.
     */
    public FeatureTableControl(Composite parent, FeatureReader fReader, int resPerPage) {        
        this.fType = fReader.getFeatureType();
        this.fReader = fReader;
        this.pageSize = resPerPage;
        features = FeatureCollections.newCollection();
        
        createTableControl( parent );
        //initial page
        next();
    }
    
    /**
     * 
     * Sets the number of features viewed in the table per page.
     * 
     * @param resPerPage positive integer.
     */
    public void setPageSize(int resPerPage) {
        this.pageSize = resPerPage;
    }
    
    /**
     * 
     * Returns the number of features viewed in the table per page.
     * 
     * @return positive integer.
     */
    public int getPageSize() {
        return pageSize;
    }
    
    /**
     * 
     * Returns the control representing the table control.
     * 
     * @return The internal table viewer control.
     */
    public Control getControl() {
        return book;
    }
    
    /**
     * Creates the table control.
     * 
     * @param parent The to be parent of the control.  
     */
    public void createTableControl(Composite parent) {
        book = new PageBook( parent, SWT.NONE );        
        message = new Text( book, SWT.NONE );
        createTableViewer(book);
        book.showPage( book );
        
        first= new Action("first") { //$NON-NLS-1$
            public void run() {
                reset();
            }
        };
        Policy.initAction(first, "action.first"); //$NON-NLS-1$
        first.setEnabled( false );
        prev = new Action("prev") { //$NON-NLS-1$
            public void run() {
                // not yet implemented
            }
        };
        Policy.initAction(prev, "action.prev"); //$NON-NLS-1$
        prev.setEnabled( false );
        
        next= new Action("next") { //$NON-NLS-1$
            public void run() {
                next();
            }
        };
        Policy.initAction(next, "action.next"); //$NON-NLS-1$
        next.setEnabled( false );
        
        last = new Action("last") { //$NON-NLS-1$
            public void run() {
                reset();
            }
        };
        Policy.initAction( last, "action.last"); //$NON-NLS-1$
        last.setEnabled( false );
        
    }
    
    /**
     * Moves the control to the next page of features. This method should
     * only be called if the control was created from a <code>org.geotools.data.FeatureReader</code>.
     *
     */
    public void next() {
        //if the control was created directory from a feature collection
        // fReader will be null
        //TODO: Throw an exception here instead ??
        if (fReader != null) {
	        nextInternal();
        }
    }
    
    /**
     * Moves the control to the next page of features.
     *
     */
    protected void nextInternal() {
      
	    //clear the current features
	    features.clear();
	    
	    try {
	        //read in the new features
	        for (int i = 0; i < pageSize && fReader.hasNext(); i++) {
	            features.add(fReader.next());
	        }
	    }
	    catch(Exception e) {
	        e.printStackTrace();
	    }
	    
    }
    
    /**
     * Updates the table control with the current set of features.
     * <p>
     * This method will ensure that the column information gets updated
     * </p>
     */
    public void update() {
        createTableViewer(book);
        tableViewer.setInput(features);
        try {
            next.setEnabled( fReader != null && fReader.hasNext() );
        } catch (IOException e) {
            next.setEnabled( false );
        }
        first.setEnabled( results != null  );        
    }
    
    /**
     * 
     * Creates the table control itself.
     * 
     * @param parent
     */
    protected void createTableViewer(Composite parent) {
        int style = SWT.MULTI ; //| SWT.BORDER | SWT.V_SCROLL | SWT.H_SCROLL;
        if( tableViewer != null ) {
            tableViewer.getControl().dispose();
            tableViewer = null;
        }
	    Table table = new Table(parent, style);
	      
          TableColumn tableColumn = new TableColumn(table, SWT.CENTER|SWT.BORDER);
          tableColumn.setText("FID"); //$NON-NLS-1$
          tableColumn.setWidth(100);
         if( fType != null ){
              for (int i = 0; i < fType.getAttributeCount(); i++) {
                  AttributeType aType = fType.getAttributeType(i);
                  tableColumn = new TableColumn(table,SWT.CENTER|SWT.BORDER);
                  if (aType.isGeometry()){
                      // jg: wot is this maddness? jd: paul said so
                      tableColumn.setText("GEOMETRY"); //$NON-NLS-1$
                  }
                  else tableColumn.setText(aType.getName());
                  tableColumn.setWidth(100);
              }
         }
         table.setHeaderVisible(true);
         FeatureTableProvider ftp = new FeatureTableProvider();
         
         tableViewer = new TableViewer(table);
         tableViewer.setContentProvider(ftp);
         tableViewer.setLabelProvider(ftp);
         tableViewer.setInput(features);       
         book.showPage(tableViewer.getControl());
    }
   
    class FeatureTableProvider extends LabelProvider 
      implements ITableLabelProvider, IStructuredContentProvider {

        /**
         * 
         * Returns the elemetns for the viewer. Input to this method should be 
         * at type of <code>org.geotools.feature.FeatureCollection</code> 
         * 
         * @see org.eclipse.jface.viewers.IStructuredContentProvider#getElements(java.lang.Object)
         * @param inputElement A <code>org.geotools.feature.FeatureCollection.</code>
         * @return An array of <code>org.geotools.feature.Feature</code>
         */
        public Object[] getElements( Object inputElement ) {
            FeatureCollection collection = (FeatureCollection)inputElement;
            return collection.toArray();
        }
        
        /**
         * 
         * Does nothing.
         * 
         * @see org.eclipse.jface.viewers.IContentProvider#inputChanged(org.eclipse.jface.viewers.Viewer, java.lang.Object, java.lang.Object)
         * @param viewer
         * @param oldInput
         * @param newInput
         */
        public void inputChanged( Viewer viewer, Object oldInput, Object newInput ) {
            //do nothing.
        }

        /**
         * 
         * Does nothing.
         * 
         * @see org.eclipse.jface.viewers.ITableLabelProvider#getColumnImage(java.lang.Object, int)
         * @param element
         * @param columnIndex
         * @return null
         */
        public Image getColumnImage( Object element, int columnIndex ) {
            return null;
        }

        /**
         *
         * Returns the value of the column / feature attribute.
         * 
         * @see org.eclipse.jface.viewers.ITableLabelProvider#getColumnText(java.lang.Object, int)
         * @param element the array of feature attributes.
         * @param columnIndex the column index / feature attribute.
         * @return the string representation of the feature attribute, except
         * for attributes of type Geometry in which a string representing the 
         * type of Geometry is returned.
         */
        public String getColumnText( Object element, int columnIndex ) {
            Feature f = (Feature)element;
            if (columnIndex == 0) return f.getID();
            
            AttributeType at = fType.getAttributeType(columnIndex-1);
            if (at.isGeometry()) {
                String s = 
                    f.getAttribute(columnIndex-1).getClass().getName();
                return s.substring(s.lastIndexOf('.')+1);
            }
            return f.getAttribute(columnIndex-1).toString();        
            
        }
    }

    /**
     * 
     * Does nothing.
     * 
     * @see org.eclipse.ui.IWorkbenchPart#setFocus()
     *
     */
    public void setFocus() {
        //do nothing.
    }
    /** Contents of the current page of features 
     * @return */
    public FeatureCollection getFeatures() {
        return features;
    }
    /** Set up for a one time only paged access */
    public void setFeatures( FeatureReader reader ) {
        this.fReader = reader;
        this.fType = reader.getFeatureType();
        features = FeatureCollections.newCollection();
        next(); // load the first page into the collection
        update();
    }
    /** Set up for a single page of content */
    public void setFeatures( FeatureCollection features ) {
        this.features = features;
        if( !features.isEmpty() ){
            fType = features.features().next().getFeatureType();
        }
        else {
            fType = null;
        }
        //this.pageSize = features.size();
        this.fReader = null;
        update();
    }
    /**
     * Set up for pages access with a next and a reset button.
     * 
     * @param results
     */
    public void setFeatures( FeatureResults results) {
        this.results = results;
        reset();
        update();
    }
    /** Reset to start of FeatureResutls */
    public void reset() {
        int count = pageSize+1;
        try {
            count = results.getCount();
//            System.out.println("results count " + count );
        } catch (IOException e) {            
//            System.out.println("results count " + e );
            e.printStackTrace();            
        }
        
        if (count < this.pageSize ){            
            try {
                setFeatures( results.collection() );
            } catch (IOException e1) {
                e1.printStackTrace();
                message( e1.getMessage() );
            }
        }
        else {
            try {
                setFeatures( results.reader() );
            } catch (IOException e1) {
                e1.printStackTrace();
                message( e1.getMessage());
            }                
        }
    }    
    /**
     * Don't display nothing :-)
     */
    public void clear(){
        fType = null;
        features = null;
        fReader = null;
        this.results = null;
    }
    public void message( String text ){
        message.setText( text );
        book.showPage( message );
        clear();
    }
}