/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.ui.operations;

import java.io.IOException;
import java.util.Iterator;

import net.refractions.udig.core.IBlockingAdaptable;
import net.refractions.udig.internal.ui.Images;
import net.refractions.udig.internal.ui.UiPlugin;
import net.refractions.udig.internal.ui.operations.OperationCategory;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;

/**
 * Creates an Action that runs an operation in a background thread when triggered.
 * 
 * @author jeichar
 * @since 0.3
 */
public class OpAction extends Action implements ISelectionListener {
    IConfigurationElement configElem;
    IStructuredSelection selection;
    private IOp operation;
    private RunJob runJob;
	private String targetClass;
    OperationCategory category;
    String menuPath;
  
    /**
     * Subclasses must have the same constructor signature as this constructor,
     * {@linkplain OperationMenuFactory} creates OpActions using this
     * constructor.
     * 
     * Construct <code>OpAction</code>.
     *
     */
    public OpAction( IConfigurationElement element) {
        this.configElem = element;
        setText(element.getAttribute("name")); //$NON-NLS-1$
        String icon=element.getAttribute("icon"); //$NON-NLS-1$
        if( icon !=null )
        	setImageDescriptor(Images.getDescriptor(icon));
        String tooltip=element.getAttribute("tooltip"); //$NON-NLS-1$
        if( tooltip!=null )
        	setToolTipText(tooltip);
        this.targetClass=element.getAttribute("targetClass"); //$NON-NLS-1$
        this.menuPath=element.getAttribute("menuPath"); //$NON-NLS-1$
        this.
        runJob=new RunJob(getText());
    }
    /**
     * Since the target object may not be the object that the operation actually operates on, the
     * getOperationTarget() finds the real object and returns it or null if for some reason the
     * operation can be performed on the target.
     * <p>
     * Example:
     * </p>
     * A FeatureType readonly operation is called on a IResolve. The getOperationTarget would
     * resolve(FeatureSource.class) and return the Feature Source for the operation.
     * 
     * @param target the object the action is called on.
     * @param definition the Configuration element definition of the operation.
     * @param monitor The progress monitor for the operation job.
     * @return
     * @throws IOException 
     */
    public Object getOperationTarget( Object obj, IProgressMonitor monitor ) throws IOException{
		Class< ? > target=null;
		try {
			target=obj.getClass().getClassLoader().loadClass(targetClass);
		} catch (ClassNotFoundException e) {
			//do nothing.
            UiPlugin.log( "This exception is bad.  The framework should not allow this to be reached.",e); //$NON-NLS-1$
			return null;
		}
			if( target.isAssignableFrom(obj.getClass()) )
				return obj;
		if (Platform.getAdapterManager().hasAdapter(obj, targetClass) )
			return Platform.getAdapterManager().getAdapter(obj,target);
		IBlockingAdaptable blockingAdaptable=getBlockingAdapter(obj);
		if( blockingAdaptable!=null && blockingAdaptable.canAdaptTo(target))
			return blockingAdaptable.getAdapter(target, monitor);
		if (obj instanceof IAdaptable) {
			IAdaptable adaptable = (IAdaptable) obj;
			return adaptable.getAdapter(target);
		}
		return null;
    }

    public void runWithEvent( Event event ) {
    	runJob.display=event.display;
        runJob.schedule();
    }

    IOp getOperation() throws CoreException {
        if (operation == null) {
            operation = (IOp) configElem.createExecutableExtension("class"); //$NON-NLS-1$
        }

        return operation;
    }

    class RunJob extends Job{
    	Display display;
        /**
         * Construct <code>RunJob</code>.
         *
         * @param name
         */
        public RunJob( String name ) {
            super(name);
        }

        protected IStatus run( IProgressMonitor monitor ) {
            try {
                IOp op = getOperation();
                
                for( Iterator iter = selection.iterator(); iter.hasNext(); ) {
                    @SuppressWarnings("unchecked") Object entry = iter.next(); //$NON-NLS-1$
                    try {
                        Object target = getOperationTarget(entry, monitor);
                        op.op(display,target, monitor);
                    } catch (Throwable e) {
                        UiPlugin.log( null, e);
                    }
                }
            } catch (Throwable e) {
                UiPlugin.log( null, e);
            }
            return Status.OK_STATUS;
        }

    }

	/**
	 * Determines whether the current operation can operate on the object.
	 * @return true if obj can be used as the operation input.
	 */
	public boolean isValid(Object obj) {
		Class< ? > target=null;
		try {
			target=obj.getClass().getClassLoader().loadClass(targetClass);
		} catch (Throwable e) {
			//do nothing this is ok.
			return false;
		}
		if( target.isAssignableFrom(obj.getClass()) )
			return true;
		
		//see if platform can adapt the object
		if (Platform.getAdapterManager().hasAdapter(obj, targetClass) )
			return true;
		
		// see if the object is a blocking adaptable object and can adapt to the correct class type
		IBlockingAdaptable blockingAdaptable=getBlockingAdapter(obj);
		if( blockingAdaptable!=null && blockingAdaptable.canAdaptTo(target))
			return true;
		
		// see if the object is adaptable and can adapt to the class type.
		IAdaptable adaptable=getAdaptable(obj);
		if( adaptable!=null ){
			return adaptable.getAdapter(target)!=null;
		}
		return false;
	}

	private IAdaptable getAdaptable(Object obj) {
		if( obj instanceof IAdaptable ){
			return (IAdaptable) obj;
		}
		if (Platform.getAdapterManager().hasAdapter(obj, IAdaptable.class.getSimpleName()) )
			return (IAdaptable) Platform.getAdapterManager().getAdapter(obj, IAdaptable.class);
		return null;
	}

	private IBlockingAdaptable getBlockingAdapter(Object obj) {
		if( obj instanceof IBlockingAdaptable ){
			return (IBlockingAdaptable) obj;
		}
		if (Platform.getAdapterManager().hasAdapter(obj, IBlockingAdaptable.class.getName()) )
			return (IBlockingAdaptable) Platform.getAdapterManager().getAdapter(obj, IBlockingAdaptable.class);
		if (obj instanceof IAdaptable) {
			IAdaptable adapter = (IAdaptable) obj;
			return (IBlockingAdaptable) adapter.getAdapter(IBlockingAdaptable.class);
		}
		return null;
	}

	/**
	 * @see org.eclipse.ui.ISelectionListener#selectionChanged(org.eclipse.ui.IWorkbenchPart, org.eclipse.jface.viewers.ISelection)
	 */
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		if (selection instanceof IStructuredSelection) {
			boolean enabled=false;
			IStructuredSelection structured = (IStructuredSelection) selection;
			for (Iterator iter = structured.iterator(); iter.hasNext();) {
				@SuppressWarnings("unchecked") Object obj = iter.next(); //$NON-NLS-1$
				if ( isValid(obj) ){
					enabled=true;
					break;
				}
			}
			setEnabled(enabled);
            if( category!=null)
                category.enablementChanged();
			this.selection=structured;
		}
	}

    public void setCategory( OperationCategory category ) {
        this.category = category;
    }
    public String getMenuPath() {
        return menuPath;
    }
    
    @Override
    public String toString() {
        return getMenuPath()+"/"+getText(); //$NON-NLS-1$
    }
}
