/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.ui.palette;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;

import net.refractions.udig.internal.ui.Policy;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

/**
 * ColourPaletteMasterList
 * 
 * Singleton class that holds 3 ColourPalette instances
 * populated with colour scheme data from xml files.
 * XML files named <type>Palettes.xml and <type>Suitability.xml
 * Where <type> is either qualitative, selective or diverging
 * 
 * Used to set Default Map colours and used when selecting new palettes
 * using Map>Properties UI
 * 
 * @author ptozer
 */
public  class ColourPaletteMasterList {
	private static ColourPalette qualitativePalette;
	private static ColourPalette selectivePalette;
	private static ColourPalette divergingPalette;
	static private ColourPaletteMasterList _instance = null;
	
	public static final String QUALITATIVE_TYPE =  "qualitative"; //$NON-NLS-1$
	public static final String SELECTIVE_TYPE =  "selective"; //$NON-NLS-1$
	public static final String DIVERGING_TYPE =  "diverging"; //$NON-NLS-1$
	private static final String DEFAULT_COLOUR_SCHEME_QUALITATIVE = "Dark2"; //$NON-NLS-1$
	
	private static HashMap<Integer, String> qualitativeOrderByLayerNumberMap;
	private static HashMap<Integer, String> selectiveOrderByLayerNumberMap;
	private static HashMap<Integer, String> divergingOrderByLayerNumberMap;
	
	/**
	 * constructor
	 */
	protected ColourPaletteMasterList() {
		super();
		qualitativePalette = new ColourPalette(QUALITATIVE_TYPE);
		selectivePalette = new ColourPalette(SELECTIVE_TYPE);
		divergingPalette = new ColourPalette(DIVERGING_TYPE);
		parseXML(QUALITATIVE_TYPE);
		initHashmaps();
		//TODO add selective and diverging
		//parseXML(SELECTIVE_TYPE);
		//parseXML(DIVERGING_TYPE);
	}
	
	/**
	 * 
	 */
	private void initHashmaps() {
		// TODO Auto-generated method stub
		qualitativeOrderByLayerNumberMap = new HashMap<Integer, String>();
		selectiveOrderByLayerNumberMap= new HashMap<Integer, String>();
		divergingOrderByLayerNumberMap= new HashMap<Integer, String>();

		qualitativeOrderByLayerNumberMap.put(0,"a"); //$NON-NLS-1$
		qualitativeOrderByLayerNumberMap.put(1,"a,b"); //$NON-NLS-1$
		qualitativeOrderByLayerNumberMap.put(2,"a,b,c"); //$NON-NLS-1$
		qualitativeOrderByLayerNumberMap.put(3,"a,b,c,d"); //$NON-NLS-1$
		qualitativeOrderByLayerNumberMap.put(4,"a,b,c,d,e"); //$NON-NLS-1$
		qualitativeOrderByLayerNumberMap.put(5,"a,b,c,d,e,f"); //$NON-NLS-1$
		qualitativeOrderByLayerNumberMap.put(6,"a,b,c,d,e,f,g"); //$NON-NLS-1$
		qualitativeOrderByLayerNumberMap.put(7,"a,b,c,d,e,f,g,h"); //$NON-NLS-1$
     
        selectiveOrderByLayerNumberMap.put(0,"c"); //$NON-NLS-1$
        selectiveOrderByLayerNumberMap.put(1,"c,f"); //$NON-NLS-1$
        selectiveOrderByLayerNumberMap.put(2,"c,f,i"); //$NON-NLS-1$
        selectiveOrderByLayerNumberMap.put(3,"b,e,g,j"); //$NON-NLS-1$
        selectiveOrderByLayerNumberMap.put(4,"b,e,g,i,k"); //$NON-NLS-1$
        selectiveOrderByLayerNumberMap.put(5,"b,d,f,g,i,k"); //$NON-NLS-1$
        selectiveOrderByLayerNumberMap.put(6,"b,d,f,g,h,j,l"); //$NON-NLS-1$
        selectiveOrderByLayerNumberMap.put(7,"a,c,d,f,g,h,j,l"); //$NON-NLS-1$
        selectiveOrderByLayerNumberMap.put(8,"a,c,d,f,h,g,j,k,l"); //$NON-NLS-1$
        
        divergingOrderByLayerNumberMap.put(0,"e"); //$NON-NLS-1$
        divergingOrderByLayerNumberMap.put(1,"e,h"); //$NON-NLS-1$
        divergingOrderByLayerNumberMap.put(2,"e,h,k"); //$NON-NLS-1$
        divergingOrderByLayerNumberMap.put(3,"c,f,j,m"); //$NON-NLS-1$
        divergingOrderByLayerNumberMap.put(4,"c,f,h,j,m"); //$NON-NLS-1$
        divergingOrderByLayerNumberMap.put(5,"b,e,g,i,k,n"); //$NON-NLS-1$
        divergingOrderByLayerNumberMap.put(6,"b,e,g,h,i,k,n"); //$NON-NLS-1$
        divergingOrderByLayerNumberMap.put(7,"b,d,f,g,i,j,l,n"); //$NON-NLS-1$
        divergingOrderByLayerNumberMap.put(8,"b,d,f,g,h,i,j,l,n"); //$NON-NLS-1$

	}

	
	/**
	    * @return The unique instance of this class.
	    */
	    static public ColourPaletteMasterList instance() {
	      if(null == _instance) {
	      	synchronized(ColourPaletteMasterList.class) {
	      		if (null == _instance) {
	      			_instance = new ColourPaletteMasterList();
	      		}
	      	}
	      }
	      return _instance;
	   }
	  
        static public ColourPalette getColourPaletteByTypeName(String type){
            if(type.equalsIgnoreCase(QUALITATIVE_TYPE)){
                 return qualitativePalette;
            }else if(type.equalsIgnoreCase(SELECTIVE_TYPE)){
                    
                return selectivePalette;
            } else if(type.equalsIgnoreCase(DIVERGING_TYPE)){
                  return divergingPalette; 
            }
            return null;
        }
	    /**
	     * Returns the string of colour letters available to the map with the
	     * given number of layers.
	     * @param type
	     * @param numberOfLayers
	     * @return
	     */
	   public String[] getColourLetters(String type, int numberOfLayers){
	   		String [] letters = null;
	   		//if number of layers is greater than the available colours, return largest choice available
	   		//TODO handle other 2 types of colour palettes
	   		
	   		if(type.equalsIgnoreCase(QUALITATIVE_TYPE)){
	   			if(numberOfLayers >= qualitativeOrderByLayerNumberMap.size()){
	   				numberOfLayers = qualitativeOrderByLayerNumberMap.size() - 1 ;
	   			}
	   			letters = new String[numberOfLayers];
	   			String s = qualitativeOrderByLayerNumberMap.get(numberOfLayers);
	   			letters = s.split(","); //$NON-NLS-1$
	   		} else if(type.equalsIgnoreCase(SELECTIVE_TYPE)){
                 if(numberOfLayers >= selectiveOrderByLayerNumberMap.size()){
                     numberOfLayers = selectiveOrderByLayerNumberMap.size() - 1 ;
                 }
                letters = new String[numberOfLayers];
                String s = selectiveOrderByLayerNumberMap.get(numberOfLayers);
                letters = s.split(","); //$NON-NLS-1$
            }else if(type.equalsIgnoreCase(DIVERGING_TYPE)){
                if(numberOfLayers >= divergingOrderByLayerNumberMap.size()){
                    numberOfLayers = divergingOrderByLayerNumberMap.size() - 1 ;
                }
               letters = new String[numberOfLayers];
               String s = divergingOrderByLayerNumberMap.get(numberOfLayers);
               letters = s.split(","); //$NON-NLS-1$
           }
	   		
	   		return letters;
	   }
	   
	   public String getColourLetterByLayerNumber(String type, int layerNumber){
	   		//if layerNumber is bigger than the total number of colours available,
	   		//pick one of them
	   	if(type.equalsIgnoreCase(QUALITATIVE_TYPE)){
	   		if(layerNumber >= qualitativeOrderByLayerNumberMap.size()){
	   			layerNumber = layerNumber % qualitativeOrderByLayerNumberMap.size();
	   		}
	   		return ((qualitativeOrderByLayerNumberMap.get(layerNumber)).split(","))[layerNumber]; //$NON-NLS-1$
	   	}
	   	return null;
	   }
	    
	   public ColourScheme getDefaultColourScheme(String type){
		   	ColourScheme scheme = null;	
		   	if(type.equalsIgnoreCase(QUALITATIVE_TYPE)){
		   		scheme =  qualitativePalette.getColourSchemeByName(DEFAULT_COLOUR_SCHEME_QUALITATIVE);
		   	}else if(type.equalsIgnoreCase(SELECTIVE_TYPE)){
		   			
		   		
		   	} else if(type.equalsIgnoreCase(DIVERGING_TYPE)){
		   			
		   	}
		   	return scheme;
	   }
	
	/**
	 * <p>Parses xml files for colour schemes (rgb values) and
	 * suitability values, each from its own file<br>
	 * Files must be named by scheme type (qualitative, selective or diverging)
	 * plus "Palettes.xml" or "Suitability.xml"<br>
	 * Populates class variables: <br>
	 * ColourPalette qualitativePalette
	 * ColourPalette selectivePalette
	 * ColourPalette divergingPalette
	 * </p>
	 * 
	 */
	private void parseXML(String type) {
		SAXBuilder builder = new SAXBuilder();
		builder.setIgnoringElementContentWhitespace(true);
       
		try {
			Document palettesDoc = builder.build(getClass().getResource( type + "Palettes.xml")); //$NON-NLS-1$
			Document suitabilityDoc = builder.build(getClass().getResource( type + "Suitability.xml")); //$NON-NLS-1$
			
//			root element is <type>
			Element paletteRoot = palettesDoc.getRootElement();
			Element suitRoot = suitabilityDoc.getRootElement();
			
			//children of type are <scheme> elements
			List allPaletteChildren = paletteRoot.getChildren();
			List allSuitChildren = suitRoot.getChildren();
			
			Element schemeElement;
			Element suitabilityElement;
			
			for(int namesIndex = 0; namesIndex < allPaletteChildren.size();namesIndex++) {
				//get the name of the scheme
				schemeElement = (Element)(allPaletteChildren.get(namesIndex));
				String schemeName  = schemeElement.getChild("schemeName").getText(); //$NON-NLS-1$

				//get all colours these are lettered
				List colourElements = schemeElement.getChildren();
				ColourScheme scheme;
                //make a new Scheme and give it its name
                if(type.equalsIgnoreCase(QUALITATIVE_TYPE)) {
                    scheme = new ColourScheme(qualitativePalette,schemeName);
                }else if(type.equalsIgnoreCase(SELECTIVE_TYPE)) {
                    scheme = new ColourScheme(selectivePalette,schemeName);
                }else if(type.equalsIgnoreCase(DIVERGING_TYPE)) {
                    scheme = new ColourScheme(divergingPalette,schemeName);
                } else {
                    throw new RuntimeException( Policy.bind("ColourPaletteMasterList.schema")+type+Policy.bind("ColourPaletteMasterList.notKnown")  ); //$NON-NLS-1$ //$NON-NLS-2$
                }
				
				
				suitabilityElement = (Element)(allSuitChildren.get(namesIndex));
				String suitSchemeName  = suitabilityElement.getChild("schemeName").getText(); //$NON-NLS-1$
				List suitElements = suitabilityElement.getChildren();
				
				//each colour scheme has a set of colours associated with it
				for(int colourElementIndex = 1; colourElementIndex < colourElements.size(); colourElementIndex++) {
					//each colour has a letter and an rgb value
					Element colourElement = (Element)colourElements.get(colourElementIndex);
					String letter = colourElement.getName();
					String rgbText = colourElement.getText();
					String [] rgbStringArray = rgbText.split(","); //$NON-NLS-1$
					int [] rgb = new int[3];
					rgb[0] = Integer.parseInt((rgbStringArray[0]).trim());
					rgb[1] = Integer.parseInt((rgbStringArray[1]).trim());
					rgb[2] = Integer.parseInt((rgbStringArray[2]).trim());
					//create a new scheme
					scheme.addColour(letter, rgb);
					
					Element suitElement = (Element)suitElements.get(colourElementIndex);
					String suitletter = suitElement.getName();
					String suitText = suitElement.getText();
					String [] suitStringArray = suitText.split(","); //$NON-NLS-1$
			
					scheme.addSuitability(letter, suitStringArray);

				}//end loop on colours

			}//end loop on schemes
				
		} catch (JDOMException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * @return Returns the divergingPalette.
	 */
	public ColourPalette getDivergingPalette() {
		return divergingPalette;
	}

	/**
	 * @return Returns the qualitativePalette.
	 */
	public ColourPalette getQualitativePalette() {
        
		return qualitativePalette;
	}

	/**
	 * @return Returns the selectivePalette.
	 */
	public ColourPalette getSelectivePalette() {
		return selectivePalette;
	}

}
